# Created on 2018-04-23 by hbldh <henrik.blidh@nedomkull.com>
"""
Base class for backend clients.
"""
import abc
from collections.abc import Callable
from typing import Any, Optional, Union

from bleak.args import SizedBuffer
from bleak.backends import BleakBackend, get_default_backend
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTServiceCollection
from bleak.exc import BleakError

NotifyCallback = Callable[[bytearray], None]


class BaseBleakClient(abc.ABC):
    """The Client Interface for Bleak Backend implementations to implement.

    The documentation of this interface should thus be safe to use as a reference for your implementation.

    Args:
        address_or_ble_device (`BLEDevice` or str): The Bluetooth address of the BLE peripheral to connect to or the `BLEDevice` object representing it.

    Keyword Args:
        timeout (float): Timeout for required ``discover`` call.
        disconnected_callback (callable): Callback that will be scheduled in the
            event loop when the client is disconnected. The callable must take one
            argument, which will be this client object.
    """

    def __init__(self, address_or_ble_device: Union[BLEDevice, str], **kwargs: Any):
        if isinstance(address_or_ble_device, BLEDevice):
            self.address = address_or_ble_device.address
        else:
            self.address = address_or_ble_device

        self.services: Optional[BleakGATTServiceCollection] = None

        self._timeout = kwargs["timeout"]
        self._disconnected_callback: Optional[Callable[[], None]] = kwargs.get(
            "disconnected_callback"
        )

    # NB: this is not marked as @abc.abstractmethod because that would break
    # 3rd-party backends. We might change this in the future to make it required.
    @property
    def name(self) -> str:
        """See :meth:`bleak.BleakClient.name`."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mtu_size(self) -> int:
        """Gets the negotiated MTU."""
        raise NotImplementedError

    # Connectivity methods

    def set_disconnected_callback(
        self, callback: Optional[Callable[[], None]], **kwargs: Any
    ) -> None:
        """Set the disconnect callback.
        The callback will only be called on unsolicited disconnect event.

        Set the callback to ``None`` to remove any existing callback.

        Args:
            callback: callback to be called on disconnection.

        """
        self._disconnected_callback = callback

    @abc.abstractmethod
    async def connect(self, pair: bool, **kwargs: Any) -> None:
        """Connect to the specified GATT server.

        Args:
            pair (bool): If the client should attempt to pair with the
            peripheral before connecting if it is not already paired.

            Backends that can't implement this should make an appropriate
            log message and ignore the parameter.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from the specified GATT server."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def pair(self, *args: Any, **kwargs: Any) -> None:
        """Pair with the peripheral."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def unpair(self) -> None:
        """Unpair with the peripheral."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        """Check connection status between this client and the server.

        Returns:
            Boolean representing connection status.

        """
        raise NotImplementedError()

    # I/O methods

    @abc.abstractmethod
    async def read_gatt_char(
        self, characteristic: BleakGATTCharacteristic, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to read from.

        Returns:
            (bytearray) The read data.

        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def read_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.

        Returns:
            The read data.

        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def write_gatt_char(
        self, characteristic: BleakGATTCharacteristic, data: SizedBuffer, response: bool
    ) -> None:
        """
        Perform a write operation on the specified GATT characteristic.

        Args:
            characteristic: The characteristic to write to.
            data: The data to send.
            response: If write-with-response operation should be done.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def write_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, data: SizedBuffer
    ) -> None:
        """Perform a write operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.
            data: The data to send (any bytes-like object).

        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def start_notify(
        self,
        characteristic: BleakGATTCharacteristic,
        callback: NotifyCallback,
        **kwargs: Any,
    ) -> None:
        """
        Activate notifications/indications on a characteristic.

        Implementers should call the OS function to enable notifications or
        indications on the characteristic.

        To keep things the same cross-platform, notifications should be preferred
        over indications if possible when a characteristic supports both.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def stop_notify(self, characteristic: BleakGATTCharacteristic) -> None:
        """Deactivate notification/indication on a specified characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to deactivate
                notification/indication on.

        """
        raise NotImplementedError()


def get_platform_client_backend_type() -> tuple[type[BaseBleakClient], BleakBackend]:
    """
    Gets the platform-specific :class:`BaseBleakClient` type.
    """
    backend = get_default_backend()
    match backend:
        case BleakBackend.P4ANDROID:
            from bleak.backends.p4android.client import (
                BleakClientP4Android,  # type: ignore
            )

            return (BleakClientP4Android, backend)  # type: ignore

        case BleakBackend.BLUEZ_DBUS:
            from bleak.backends.bluezdbus.client import (
                BleakClientBlueZDBus,  # type: ignore
            )

            return (BleakClientBlueZDBus, backend)  # type: ignore

        case BleakBackend.PYTHONISTA_CB:
            try:
                from bleak_pythonista import BleakClientPythonistaCB  # type: ignore

                return (BleakClientPythonistaCB, backend)  # type: ignore
            except ImportError as e:
                raise ImportError(
                    "Ensure you have `bleak-pythonista` package installed."
                ) from e

        case BleakBackend.CORE_BLUETOOTH:
            from bleak.backends.corebluetooth.client import (
                BleakClientCoreBluetooth,  # type: ignore
            )

            return (BleakClientCoreBluetooth, backend)  # type: ignore

        case BleakBackend.WIN_RT:
            from bleak.backends.winrt.client import BleakClientWinRT  # type: ignore

            return (BleakClientWinRT, backend)  # type: ignore

        case _:
            raise BleakError(f"Unsupported backend: {backend}")



import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "linux":
        assert False, "This is only available on Linux"


import asyncio
import contextlib
import logging
from typing import AsyncGenerator

from bumble.controller import Controller
from bumble.link import LocalLink
from bumble.transport import open_transport
from bumble.transport.common import Transport
from dbus_fast import BusType, Message, MessageType, Variant
from dbus_fast.aio.message_bus import MessageBus

from bleak._compat import timeout as async_timeout
from bleak.backends.bluezdbus import defs
from bleak.backends.bluezdbus.signals import MatchRules, add_match
from bleak.backends.bluezdbus.utils import assert_reply, get_dbus_authenticator

BLEAK_TEST_MANUFACTURER_ID = 0xB1EA

logger = logging.getLogger(__name__)


@contextlib.asynccontextmanager
async def open_message_bus() -> AsyncGenerator[MessageBus, None]:
    """
    Open a D-Bus message bus connection.
    """
    bus = MessageBus(bus_type=BusType.SYSTEM, auth=get_dbus_authenticator())
    try:
        await bus.connect()
        yield bus
    finally:
        bus.disconnect()
        await bus.wait_for_disconnect()


async def power_on_controller(
    bus: MessageBus, adapter_path: str, timeout: float = 2.0
) -> None:
    """
    Power on a Bluetooth controller via D-Bus.

    It may take some time for the adapter to fully configure itself after
    powering on. So we make multiple attempts until timeout is reached.
    """
    async with async_timeout(timeout):
        while True:
            try:
                reply = await bus.call(
                    Message(
                        destination=defs.BLUEZ_SERVICE,
                        path=adapter_path,
                        interface=defs.PROPERTIES_INTERFACE,
                        member="Set",
                        signature="ssv",
                        body=[defs.ADAPTER_INTERFACE, "Powered", Variant("b", True)],
                    )
                )
                assert_reply(reply)
                return
            except Exception as e:
                logger.debug("Failed to power on adapter at %s: %s", adapter_path, e)
                await asyncio.sleep(0.1)


@contextlib.asynccontextmanager
async def wait_for_new_adapter() -> (
    AsyncGenerator[tuple[MessageBus, asyncio.Future[str]], None]
):
    """
    Connect to D-Bus and wait for a new Bluetooth adapter to be added.

    Yields a Future that will be resolved with the adapter path when
    a new adapter is detected via InterfacesAdded signal.
    """
    async with open_message_bus() as bus:
        loop = asyncio.get_running_loop()
        adapter_path_future: asyncio.Future[str] = loop.create_future()

        def _on_interfaces_added(message: Message):
            if message.message_type != MessageType.SIGNAL:
                return
            if message.member != "InterfacesAdded":
                return

            obj_path, ifaces = message.body
            adapter = ifaces.get(defs.ADAPTER_INTERFACE)
            if not adapter:
                return

            # Check for our test manufacturer ID to identify the adapter
            if adapter["Manufacturer"].value != BLEAK_TEST_MANUFACTURER_ID:
                return

            if not adapter_path_future.done():
                adapter_path_future.set_result(obj_path)

        bus.add_message_handler(_on_interfaces_added)

        try:
            # Subscribe to InterfacesAdded signals
            reply = await add_match(
                bus,
                MatchRules(
                    interface=defs.OBJECT_MANAGER_INTERFACE,
                    member="InterfacesAdded",
                    arg0path="/org/bluez/",
                ),
            )
            assert_reply(reply)

            # Yield the future and bus for the caller to use
            yield bus, adapter_path_future
        finally:
            bus.remove_message_handler(_on_interfaces_added)


def _clear_bit(flags: bytes, bit_pos: int) -> bytes:
    int_flags = int.from_bytes(flags, byteorder="little")
    int_flags &= ~(1 << bit_pos)
    return int_flags.to_bytes(len(flags), byteorder="little")


@contextlib.asynccontextmanager
async def open_bluez_bluetooth_controller_link(
    hci_transport_name: str,
) -> AsyncGenerator[LocalLink, None]:
    """
    Open a local link (virtual RF connection) to a bumble Bluetooth
    controller that is connected to BlueZ.
    """
    async with wait_for_new_adapter() as (bus, adapter_path_future):
        # Open a HCI transport connected to the OS
        async with await open_transport(hci_transport_name) as hci_transport:
            # Local link to create virtual RF connection between multiple Bluetooth Controllers
            link = LocalLink()

            # Bluetooth controller that BlueZ can connect to.
            # (This will register itself to the link.)
            bluez_controller = Controller(
                "BLEAK-TEST-BLUEZ",
                host_source=hci_transport.source,
                host_sink=hci_transport.sink,
                link=link,
            )
            bluez_controller.manufacturer_name = BLEAK_TEST_MANUFACTURER_ID

            # HACK: Work around Bumble missing feature combined with Linux kernel
            # requirement. https://github.com/google/bumble/issues/841
            #
            # According to the Bluetooth spec:
            #
            #   C24: [HCI_LE_Enhanced_Connection_Complete event is] Mandatory if
            #   the LE Controller supports Connection State and either LE Feature (LL
            #   Privacy) or LE Feature (Extended Advertising) is supported, otherwise optional if
            #   the LE Controller supports Connection State, otherwise excluded.
            #
            # And the Linux kernel enforces this in hci_le_create_conn_sync().
            # It will get a timeout if one of these features is enabled and the
            # Enhanced Connection Complete event is not sent.
            #
            # However, Bumble (as of 0.0.220) always sends HCI_LE_Connection_Complete_Event
            # in response to HCI_LE_Create_Connection_Command even when it should
            # be sending HCI_LE_Enhanced_Connection_Complete_Event.
            #
            # For now, we can work around the issue by disabling LL Privacy and
            # Extended Advertising features in the BlueZ controller.
            #
            # Ideally, this should be fixed in Bumble.

            bluez_controller.le_features = _clear_bit(
                bluez_controller.le_features, 6  # LL Privacy
            )
            bluez_controller.le_features = _clear_bit(
                bluez_controller.le_features, 12  # Extended Advertising
            )

            # Wait up to 5 seconds for the new adapter to appear via InterfacesAdded
            adapter_path = await asyncio.wait_for(adapter_path_future, timeout=5.0)
            logger.info(f"New adapter detected at {adapter_path}")

            # Ensure controller is powered on. This also ensures that BlueZ has fully
            # initialized the adapter and it is ready for use.
            await power_on_controller(bus, adapter_path)

            # We dont need the bus anymore. We have done everything needed with it.
            # Bleak will open its own D-Bus connection.
            bus.disconnect()
            await bus.wait_for_disconnect()

            # Yield the local link for use in tests
            yield link


@contextlib.asynccontextmanager
async def open_transport_with_bluez_vhci() -> AsyncGenerator[Transport, None]:
    """
    Create a bumble HCI Transport connected to BlueZ via the vhci driver and connect
    a Bluetooth controller for a peripheral device to it.
    """
    async with open_bluez_bluetooth_controller_link("vhci") as local_link:
        peripheral_controller = Controller("BLEAK-TEST-PERIPHERAL", link=local_link)
        yield Transport(peripheral_controller, peripheral_controller)



"""Top-level package for bleak."""

from __future__ import annotations

from bleak.args import SizedBuffer

__author__ = """Henrik Blidh"""
__email__ = "henrik.blidh@gmail.com"

import asyncio
import functools
import inspect
import logging
import os
import sys
import uuid
from collections.abc import AsyncGenerator, Awaitable, Callable, Iterable
from types import TracebackType
from typing import Any, Literal, Optional, TypedDict, Union, cast, overload

from bleak._compat import Never, Self, Unpack, assert_never
from bleak._compat import timeout as async_timeout
from bleak.args.bluez import BlueZNotifyArgs, BlueZScannerArgs
from bleak.args.corebluetooth import CBScannerArgs, CBStartNotifyArgs
from bleak.args.winrt import WinRTClientArgs
from bleak.backends import BleakBackend
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.client import BaseBleakClient, get_platform_client_backend_type
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import (
    AdvertisementData,
    AdvertisementDataCallback,
    AdvertisementDataFilter,
    BaseBleakScanner,
    get_platform_scanner_backend_type,
)
from bleak.backends.service import BleakGATTServiceCollection
from bleak.exc import BleakCharacteristicNotFoundError, BleakError
from bleak.uuids import normalize_uuid_str

_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())
if bool(os.environ.get("BLEAK_LOGGING", False)):
    FORMAT = "%(asctime)-15s %(name)-8s %(threadName)s %(levelname)s: %(message)s"
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(fmt=FORMAT))
    _logger.addHandler(handler)
    _logger.setLevel(logging.DEBUG)


# prevent tasks from being garbage collected
_background_tasks: set[asyncio.Task[None]] = set()


class BleakScanner:
    """
    Interface for Bleak Bluetooth LE Scanners.

    The scanner will listen for BLE advertisements, optionally filtering on advertised services or
    other conditions, and collect a list of :class:`BLEDevice` objects. These can subsequently be used to
    connect to the corresponding BLE server.

    A :class:`BleakScanner` can be used as an asynchronous context manager in which case it automatically
    starts and stops scanning.

    Args:
        detection_callback:
            Optional function that will be called each time a device is
            discovered or advertising data has changed.
        service_uuids:
            Optional list of service UUIDs to filter on. Only advertisements
            containing this advertising data will be received. Required on
            macOS >= 12.0, < 12.3 (unless you create an app with ``py2app``).
        scanning_mode:
            Set to ``"passive"`` to avoid the ``"active"`` scanning mode.
            Passive scanning is not supported on macOS! Will raise
            :class:`BleakError` if set to ``"passive"`` on macOS.
        bluez:
            Dictionary of arguments specific to the BlueZ backend.
        cb:
            Dictionary of arguments specific to the CoreBluetooth backend.
        backend:
            Used to override the automatically selected backend (i.e. for a
            custom backend).
        **kwargs:
            Additional args for backwards compatibility.

    .. tip:: The first received advertisement in ``detection_callback`` may or
        may not include scan response data if the remote device supports it.
        Be sure to take this into account when handing the callback. For example,
        the scan response often contains the local name of the device so if you
        are matching a device based on other data but want to display the local
        name to the user, be sure to wait for ``adv_data.local_name is not None``.

    .. versionchanged:: 0.15
        ``detection_callback``, ``service_uuids`` and ``scanning_mode`` are no longer keyword-only.
        Added ``bluez`` parameter.

    .. versionchanged:: 0.18
        No longer is alias for backend type and no longer inherits from :class:`BaseBleakScanner`.
        Added ``backend`` parameter.
    """

    def __init__(
        self,
        detection_callback: Optional[AdvertisementDataCallback] = None,
        service_uuids: Optional[list[str]] = None,
        scanning_mode: Literal["active", "passive"] = "active",
        *,
        bluez: BlueZScannerArgs = {},
        cb: CBScannerArgs = {},
        backend: Optional[type[BaseBleakScanner]] = None,
        **kwargs: Any,
    ) -> None:
        PlatformBleakScanner, backend_id = (
            get_platform_scanner_backend_type()
            if backend is None
            else (backend, backend.__name__)
        )

        self._backend = PlatformBleakScanner(
            detection_callback,
            service_uuids,
            scanning_mode,  # type: ignore
            bluez=bluez,
            cb=cb,
            **kwargs,
        )  # type: ignore
        self._backend_id = backend_id

    @property
    def backend_id(self) -> BleakBackend | str:
        """
        Gets the identifier of the backend in use.

        The value is one of the :class:`BleakBackend` enum values in case of
        built-in backends, or a string identifying a custom backend.

        .. versionadded:: 2.0
        """
        return self._backend_id

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        await self.stop()

    async def start(self) -> None:
        """
        Start scanning for devices.

        Raises:
            BleakBluetoothNotAvailableError:
                if Bluetooth is not currently available

        .. versionchanged:: 2.0
            Now raises :class:`BleakBluetoothNotAvailableError` instead of :class:`BleakError`
            when Bluetooth is not currently available.
        """
        await self._backend.start()

    async def stop(self) -> None:
        """Stop scanning for devices"""
        await self._backend.stop()

    async def advertisement_data(
        self,
    ) -> AsyncGenerator[tuple[BLEDevice, AdvertisementData], None]:
        """
        Yields devices and associated advertising data packets as they are discovered.

        .. note::
            Ensure that scanning is started before calling this method.

        Returns:
            An async iterator that yields tuples (:class:`BLEDevice`, :class:`AdvertisementData`).

        .. versionadded:: 0.21
        """
        devices: asyncio.Queue[tuple[BLEDevice, AdvertisementData]] = asyncio.Queue()

        unregister_callback = self._backend.register_detection_callback(
            lambda bd, ad: devices.put_nowait((bd, ad))
        )
        try:
            while True:
                yield await devices.get()
        finally:
            unregister_callback()

    class ExtraArgs(TypedDict, total=False):
        """
        Keyword args from :class:`~bleak.BleakScanner` that can be passed to
        other convenience methods.
        """

        service_uuids: list[str]
        """
        Optional list of service UUIDs to filter on. Only advertisements
        containing this advertising data will be received. Required on
        macOS >= 12.0, < 12.3 (unless you create an app with ``py2app``).
        """
        scanning_mode: Literal["active", "passive"]
        """
        Set to ``"passive"`` to avoid the ``"active"`` scanning mode.
        Passive scanning is not supported on macOS! Will raise
        :class:`BleakError` if set to ``"passive"`` on macOS.
        """
        bluez: BlueZScannerArgs
        """
        Dictionary of arguments specific to the BlueZ backend.
        """
        cb: CBScannerArgs
        """
        Dictionary of arguments specific to the CoreBluetooth backend.
        """
        backend: type[BaseBleakScanner]
        """
        Used to override the automatically selected backend (i.e. for a
            custom backend).
        """
        adapter: str | None
        """
        Name of adapter to use (BlueZ specific), e.g. hci0.
        """

    @overload
    @classmethod
    async def discover(
        cls,
        timeout: float = 5.0,
        *,
        return_adv: Literal[False] = False,
        **kwargs: Unpack[ExtraArgs],
    ) -> list[BLEDevice]: ...

    @overload
    @classmethod
    async def discover(
        cls,
        timeout: float = 5.0,
        *,
        return_adv: Literal[True],
        **kwargs: Unpack[ExtraArgs],
    ) -> dict[str, tuple[BLEDevice, AdvertisementData]]: ...

    @classmethod
    async def discover(
        cls,
        timeout: float = 5.0,
        *,
        return_adv: bool = False,
        **kwargs: Unpack[ExtraArgs],
    ):
        """
        Scan continuously for ``timeout`` seconds and return discovered devices.

        Args:
            timeout:
                Time, in seconds, to scan for.
            return_adv:
                If ``True``, the return value will include advertising data.
            **kwargs:
                Additional arguments will be passed to the :class:`BleakScanner`
                constructor.

        Returns:
            The value of :attr:`discovered_devices_and_advertisement_data` if
            ``return_adv`` is ``True``, otherwise the value of :attr:`discovered_devices`.

        .. versionchanged:: 0.19
            Added ``return_adv`` parameter.
        """
        async with cls(**kwargs) as scanner:
            await asyncio.sleep(timeout)

        if return_adv:
            return scanner.discovered_devices_and_advertisement_data

        return scanner.discovered_devices

    @property
    def discovered_devices(self) -> list[BLEDevice]:
        """
        Gets list of the devices that the scanner has discovered during the scanning.

        If you also need advertisement data, use :attr:`discovered_devices_and_advertisement_data` instead.
        """
        return [d for d, _ in self._backend.seen_devices.values()]

    @property
    def discovered_devices_and_advertisement_data(
        self,
    ) -> dict[str, tuple[BLEDevice, AdvertisementData]]:
        """
        Gets a map of device address to tuples of devices and the most recently
        received advertisement data for that device.

        The address keys are useful to compare the discovered devices to a set
        of known devices. If you don't need to do that, consider using
        ``discovered_devices_and_advertisement_data.values()`` to just get the
        values instead.

        .. versionadded:: 0.19
        """
        return {d[0].address: d for d in self._backend.seen_devices.values()}

    @classmethod
    async def find_device_by_address(
        cls, device_identifier: str, timeout: float = 10.0, **kwargs: Unpack[ExtraArgs]
    ) -> Optional[BLEDevice]:
        """Obtain a ``BLEDevice`` for a BLE server specified by Bluetooth address or (macOS) UUID address.

        Args:
            device_identifier: The Bluetooth/UUID address of the Bluetooth peripheral sought.
            timeout: Optional timeout to wait for detection of specified peripheral before giving up. Defaults to 10.0 seconds.
            **kwargs: additional args passed to the :class:`BleakScanner` constructor.

        Returns:
            The ``BLEDevice`` sought or ``None`` if not detected.

        """
        device_identifier = device_identifier.lower()
        return await cls.find_device_by_filter(
            lambda d, ad: d.address.lower() == device_identifier,
            timeout=timeout,
            **kwargs,
        )

    @classmethod
    async def find_device_by_name(
        cls, name: str, timeout: float = 10.0, **kwargs: Unpack[ExtraArgs]
    ) -> Optional[BLEDevice]:
        """Obtain a ``BLEDevice`` for a BLE server specified by the local name in the advertising data.

        Args:
            name: The name sought.
            timeout: Optional timeout to wait for detection of specified peripheral before giving up. Defaults to 10.0 seconds.
            **kwargs: additional args passed to the :class:`BleakScanner` constructor.

        Returns:
            The ``BLEDevice`` sought or ``None`` if not detected.

        .. versionadded:: 0.20
        """
        return await cls.find_device_by_filter(
            lambda d, ad: ad.local_name == name,
            timeout=timeout,
            **kwargs,
        )

    @classmethod
    async def find_device_by_filter(
        cls,
        filterfunc: AdvertisementDataFilter,
        timeout: float = 10.0,
        **kwargs: Unpack[ExtraArgs],
    ) -> Optional[BLEDevice]:
        """Obtain a ``BLEDevice`` for a BLE server that matches a given filter function.

        This can be used to find a BLE server by other identifying information than its address,
        for example its name.

        Args:
            filterfunc:
                A function that is called for every BLEDevice found. It should
                return ``True`` only for the wanted device.
            timeout:
                Optional timeout to wait for detection of specified peripheral
                before giving up. Defaults to 10.0 seconds.
            **kwargs:
                Additional arguments to be passed to the :class:`BleakScanner`
                constructor.

        Returns:
            The :class:`BLEDevice` sought or ``None`` if not detected before
            the timeout.

        """
        async with cls(**kwargs) as scanner:
            try:
                async with async_timeout(timeout):
                    async for bd, ad in scanner.advertisement_data():
                        if filterfunc(bd, ad):
                            return bd
                    assert_never(cast(Never, "advertisement_data() should never stop"))
            except asyncio.TimeoutError:
                return None


def _resolve_characteristic(
    char_specifier: Union[BleakGATTCharacteristic, int, str, uuid.UUID],
    services: BleakGATTServiceCollection,
) -> BleakGATTCharacteristic:

    if isinstance(char_specifier, BleakGATTCharacteristic):
        return char_specifier

    characteristic = services.get_characteristic(char_specifier)

    if not characteristic:
        raise BleakCharacteristicNotFoundError(char_specifier)

    return characteristic


def _resolve_descriptor(
    desc_specifier: Union[BleakGATTDescriptor, int],
    services: BleakGATTServiceCollection,
) -> BleakGATTDescriptor:

    if isinstance(desc_specifier, BleakGATTDescriptor):
        return desc_specifier

    characteristic = services.get_descriptor(desc_specifier)

    if not characteristic:
        raise BleakError(f"Descriptor with handle {desc_specifier} was not found!")

    return characteristic


class BleakClient:
    """The Client interface for connecting to a specific BLE GATT server and communicating with it.

    A BleakClient can be used as an asynchronous context manager in which case it automatically
    connects and disconnects.

    How many BLE connections can be active simultaneously, and whether connections can be active while
    scanning depends on the Bluetooth adapter hardware.

    Args:
        address_or_ble_device:
            A :class:`BLEDevice` received from a :class:`BleakScanner` or a
            Bluetooth address (device UUID on macOS).
        disconnected_callback:
            Callback that will be scheduled in the event loop when the client is
            disconnected. The callable must take one argument, which will be
            this client object.
        services:
            Optional list of services to filter. If provided, only these services
            will be resolved. This may or may not reduce the time needed to
            enumerate the services depending on if the OS supports such filtering
            in the Bluetooth stack or not (should affect Windows and Mac).
            These can be 16-bit or 128-bit UUIDs.
        timeout:
            Timeout in seconds passed to the implicit ``discover`` call when
            ``address_or_ble_device`` is not a :class:`BLEDevice`. Defaults to 30.
        pair:
            Attempt to pair with the the device before connecting, if it is not
            already paired. This has no effect on macOS since pairing is initiated
            automatically when accessing a characteristic that requires authentication.
            In rare cases, on other platforms, it might be necessary to pair the
            device first in order to be able to even enumerate the services during
            the connection process.
        winrt:
            Dictionary of WinRT/Windows platform-specific options.
        backend:
            Used to override the automatically selected backend (i.e. for a
            custom backend).
        **kwargs:
            Additional keyword arguments for backwards compatibility.

    .. tip:: If you enable pairing with the ``pair`` argument, you will also
        want to extend the timeout to allow enough time for the user to find
        and enter the PIN code on the device, if required.

    .. warning:: Although example code frequently initializes :class:`BleakClient`
        with a Bluetooth address for simplicity, it is not recommended to do so
        for more complex use cases. There are several known issues with providing
        a Bluetooth address as the ``address_or_ble_device`` argument.

        1.  macOS does not provide access to the Bluetooth address for privacy/
            security reasons. Instead it creates a UUID for each Bluetooth
            device which is used in place of the address on this platform.
        2.  Providing an address or UUID instead of a :class:`BLEDevice` causes
            the :meth:`connect` method to implicitly call :meth:`BleakScanner.discover`.
            This is known to cause problems when trying to connect to multiple
            devices at the same time.

    .. versionchanged:: 0.15
        ``disconnected_callback`` is no longer keyword-only. Added ``winrt`` parameter.

    .. versionchanged:: 0.18
        No longer is alias for backend type and no longer inherits from :class:`BaseBleakClient`.
        Added ``backend`` parameter.

    .. versionchanged:: 1.0
        Added ``pair`` parameter.

    .. versionchanged:: 2.1.1
        Changed default connect timeout from 10 to 30 seconds.
    """

    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        disconnected_callback: Optional[Callable[[BleakClient], None]] = None,
        services: Optional[Iterable[str]] = None,
        *,
        timeout: float = 30,
        pair: bool = False,
        winrt: WinRTClientArgs = {},
        backend: Optional[type[BaseBleakClient]] = None,
        **kwargs: Any,
    ) -> None:
        PlatformBleakClient, backend_id = (
            get_platform_client_backend_type()
            if backend is None
            else (backend, backend.__name__)
        )

        self._backend = PlatformBleakClient(
            address_or_ble_device,
            disconnected_callback=(
                None
                if disconnected_callback is None
                else functools.partial(disconnected_callback, self)
            ),
            services=(
                None if services is None else set(map(normalize_uuid_str, services))
            ),
            timeout=timeout,
            winrt=winrt,
            **kwargs,
        )
        self._pair_before_connect = pair
        self._backend_id = backend_id

    @property
    def backend_id(self) -> BleakBackend | str:
        """
        Gets the identifier of the backend in use.

        The value is one of the :class:`BleakBackend` enum values in case of
        built-in backends, or a string identifying a custom backend.

        .. versionadded:: 2.0
        """
        return self._backend_id

    # device info

    @property
    def name(self) -> str:
        """
        Gets a human-readable name for the peripheral device.

        The name can be somewhat OS-dependent. It is usually the name provided
        by the standard Device Name characteristic, if present or the name
        provided by the advertising data. If neither is available, it will be
        a Bluetooth address separated with dashes (``-``) instead of colons
        (``:``) (or a UUID on Apple devices). It may also be possible to override
        the device name using the OS's Bluetooth settings.

        .. versionadded:: 1.1
        """
        return self._backend.name

    @property
    def address(self) -> str:
        """
        Gets the Bluetooth address of this device (UUID on macOS).
        """
        return self._backend.address

    @property
    def mtu_size(self) -> int:
        """
        Gets the negotiated MTU size in bytes for the active connection.

        Consider using :attr:`bleak.backends.characteristic.BleakGATTCharacteristic.max_write_without_response_size` instead.

        .. warning:: The BlueZ backend will always return 23 (the minimum MTU size).
            See the ``mtu_size.py`` example for a way to hack around this.

        """
        return self._backend.mtu_size

    def __str__(self) -> str:
        return f"{self.__class__.__name__}, {self.address}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}, {self.address}, {type(self._backend)}>"

    # Async Context managers

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.disconnect()

    # Connectivity methods

    async def connect(self, **kwargs: Any) -> None:
        """Connect to the specified GATT server.

        Args:
            **kwargs: For backwards compatibility - should not be used.

        .. versionchanged:: 1.0
            No longer returns ``True``. Instead, the return type is ``None``.
        """
        await self._backend.connect(self._pair_before_connect, **kwargs)

    async def disconnect(self) -> None:
        """Disconnect from the specified GATT server.

        .. versionchanged:: 1.0
            No longer returns ``True``. Instead, the return type is ``None``.
        """
        await self._backend.disconnect()

    async def pair(self, *args: Any, **kwargs: Any) -> None:
        """
        Pair with the specified GATT server.

        This method is not available on macOS. Instead of manually initiating
        paring, the user will be prompted to pair the device the first time
        that a characteristic that requires authentication is read or written.
        This method may have backend-specific additional keyword arguments.

        .. versionchanged:: 1.0
            No longer returns ``True``. Instead, the return type is ``None``.
        """
        await self._backend.pair(*args, **kwargs)

    async def unpair(self) -> None:
        """
        Unpair from the specified GATT server.

        Unpairing will also disconnect the device.

        This method is only available on Windows and Linux and will raise an
        exception on other platforms.

        .. versionchanged:: 1.0
            No longer returns ``True``. Instead, the return type is ``None``.
        """
        await self._backend.unpair()

    @property
    def is_connected(self) -> bool:
        """
        Check connection status between this client and the GATT server.

        Returns:
            Boolean representing connection status.

        """
        return self._backend.is_connected

    # GATT services methods

    @property
    def services(self) -> BleakGATTServiceCollection:
        """
        Gets the collection of GATT services available on the device.

        The returned value is only valid as long as the device is connected.

        Raises:
            BleakError: if service discovery has not been performed yet during this connection.
        """
        if not self._backend.services:
            raise BleakError("Service Discovery has not been performed yet")

        return self._backend.services

    # I/O methods

    async def read_gatt_char(
        self,
        char_specifier: Union[BleakGATTCharacteristic, int, str, uuid.UUID],
        **kwargs: Any,
    ) -> bytearray:
        """
        Perform read operation on the specified GATT characteristic.

        Args:
            char_specifier:
                The characteristic to read from, specified by either integer
                handle, UUID or directly by the BleakGATTCharacteristic object
                representing it.

        Returns:
            The read data.

        Raises:
            BleakCharacteristicNotFoundError: if a characteristic with the
                handle or UUID specified by ``char_specifier`` could not be found.
            backend-specific exceptions: if the read operation failed.
        """
        characteristic = _resolve_characteristic(char_specifier, self.services)
        return await self._backend.read_gatt_char(characteristic, **kwargs)

    async def write_gatt_char(
        self,
        char_specifier: Union[BleakGATTCharacteristic, int, str, uuid.UUID],
        data: SizedBuffer,
        response: Optional[bool] = None,
    ) -> None:
        r"""
        Perform a write operation on the specified GATT characteristic.

        There are two possible kinds of writes. *Write with response* (sometimes
        called a *Request*) will write the data then wait for a response from
        the remote device. *Write without response* (sometimes called *Command*)
        will queue data to be written and return immediately.

        Each characteristic may support one kind or the other or both or neither.
        Consult the device's documentation or inspect the properties of the
        characteristic to find out which kind of writes are supported.

        Args:
            char_specifier:
                The characteristic to write to, specified by either integer
                handle, UUID or directly by the :class:`~bleak.backends.characteristic.BleakGATTCharacteristic`
                object representing it. If a device has more than one characteristic
                with the same UUID, then attempting to use the UUID wil fail and
                a characteristic object must be used instead.
            data:
                The data to send. When a write-with-response operation is used,
                the length of the data is limited to 512 bytes. When a
                write-without-response operation is used, the length of the
                data is limited to :attr:`~bleak.backends.characteristic.BleakGATTCharacteristic.max_write_without_response_size`.
                Any type that supports the buffer protocol can be passed.
            response:
                If ``True``, a write-with-response operation will be used. If
                ``False``, a write-without-response operation will be used.
                Omitting the argument is deprecated and may raise a warning.
                If this arg is omitted, the default behavior is to check the
                characteristic properties to see if the "write" property is
                present. If it is, a write-with-response operation will be
                used. Note: some devices may incorrectly report or omit the
                property, which is why an explicit argument is encouraged.

        Raises:
            BleakCharacteristicNotFoundError: if a characteristic with the
                handle or UUID specified by ``char_specifier`` could not be found.
            backend-specific exceptions: if the write operation failed.

        .. versionchanged:: 0.21
            The default behavior when ``response=`` is omitted was changed.

        Example::

            MY_CHAR_UUID = "1234"
            ...
            await client.write_gatt_char(MY_CHAR_UUID, b"\x00\x01\x02\x03", response=True)
        """
        characteristic = _resolve_characteristic(char_specifier, self.services)

        if response is None:
            # If not specified, prefer write-with-response over write-without-
            # response if it is available since it is the more reliable write.
            # This assumes that the peripheral correctly reports the
            # characteristic properties, so doesn't work in some cases.
            response = "write" in characteristic.properties

        await self._backend.write_gatt_char(characteristic, data, response)

    async def start_notify(
        self,
        char_specifier: Union[BleakGATTCharacteristic, int, str, uuid.UUID],
        callback: Callable[
            [BleakGATTCharacteristic, bytearray], Union[None, Awaitable[None]]
        ],
        *,
        bluez: BlueZNotifyArgs = {},
        cb: CBStartNotifyArgs = {},
        **kwargs: Any,
    ) -> None:
        """
        Activate notifications/indications on a characteristic.

        Callbacks must accept two inputs. The first will be the characteristic
        and the second will be a ``bytearray`` containing the data received.

        .. code-block:: python

            def callback(sender: BleakGATTCharacteristic, data: bytearray):
                print(f"{sender}: {data}")

            client.start_notify(char_uuid, callback)

        Args:
            char_specifier:
                The characteristic to activate notifications/indications on a
                characteristic, specified by either integer handle,
                UUID or directly by the BleakGATTCharacteristic object representing it.
            callback:
                The function to be called on notification. Can be regular
                function or async function.
            bluez:
                BlueZ backend-specific arguments.
            cb:
                CoreBluetooth backend-specific arguments.

        Raises:
            BleakCharacteristicNotFoundError: if a characteristic with the
                handle or UUID specified by ``char_specifier`` could not be found.
            backend-specific exceptions: if the start notification operation failed.

        .. versionchanged:: 0.18
            The first argument of the callback is now a :class:`BleakGATTCharacteristic`
            instead of an ``int``.

        .. versionchanged:: 1.0
            Added the ``cb`` parameter.

        .. versionchanged:: 2.1
            Added the ``bluez`` parameter.
        """
        if not self.is_connected:
            raise BleakError("Not connected")

        characteristic = _resolve_characteristic(char_specifier, self.services)

        if inspect.iscoroutinefunction(callback):

            def wrapped_callback(data: bytearray) -> None:
                task = asyncio.create_task(callback(characteristic, data))
                _background_tasks.add(task)
                task.add_done_callback(_background_tasks.discard)

        else:
            wrapped_callback = functools.partial(callback, characteristic)  # type: ignore

        await self._backend.start_notify(
            characteristic, wrapped_callback, bluez=bluez, cb=cb, **kwargs
        )

    async def stop_notify(
        self, char_specifier: Union[BleakGATTCharacteristic, int, str, uuid.UUID]
    ) -> None:
        """
        Deactivate notification/indication on a specified characteristic.

        Args:
            char_specifier:
                The characteristic to deactivate notification/indication on,
                specified by either integer handle, UUID or directly by the
                BleakGATTCharacteristic object representing it.

        Raises:
            BleakCharacteristicNotFoundError: if a characteristic with the
                handle or UUID specified by ``char_specifier`` could not be found.
            backend-specific exceptions: if the stop notification operation failed.

        .. tip:: Notifications are stopped automatically on disconnect, so this
            method does not need to be called unless notifications need to be
            stopped some time before the device disconnects.
        """
        characteristic = _resolve_characteristic(char_specifier, self.services)
        await self._backend.stop_notify(characteristic)

    async def read_gatt_descriptor(
        self,
        desc_specifier: Union[BleakGATTDescriptor, int],
        **kwargs: Any,
    ) -> bytearray:
        """
        Perform read operation on the specified GATT descriptor.

        Args:
            desc_specifier:
                The descriptor to read from, specified by either integer handle
                or directly by the BleakGATTDescriptor object representing it.

        Raises:
            BleakError: if the descriptor could not be found.
            backend-specific exceptions: if the read operation failed.

        Returns:
            The read data.

        """
        descriptor = _resolve_descriptor(desc_specifier, self.services)
        return await self._backend.read_gatt_descriptor(descriptor, **kwargs)

    async def write_gatt_descriptor(
        self,
        desc_specifier: Union[BleakGATTDescriptor, int],
        data: SizedBuffer,
    ) -> None:
        """
        Perform a write operation on the specified GATT descriptor.

        Args:
            desc_specifier:
                The descriptor to write to, specified by either integer handle
                directly by the BleakGATTDescriptor object representing it.
            data:
                The data to send.

        Raises:
            BleakError: if the descriptor could not be found.
            backend-specific exceptions: if the read operation failed.

        """
        descriptor = _resolve_descriptor(desc_specifier, self.services)
        await self._backend.write_gatt_descriptor(descriptor, data)



import pytest
from bumble.device import Device

from bleak import BleakClient
from bleak.backends import BleakBackend, get_default_backend
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    find_ble_device,
)


@pytest.mark.skipif(
    get_default_backend() != BleakBackend.CORE_BLUETOOTH,
    reason="pairing is not possible on CoreBluetooth",
)
async def test_pairing_unavailable(bumble_peripheral: Device):
    """Check if pairing on CoreBluetooth raises an error."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    client = BleakClient(device)
    with pytest.raises(NotImplementedError):
        await client.pair()
    with pytest.raises(NotImplementedError):
        await client.unpair()


# TODO: Add tests for pairing



"""
Example showing how to use BleakClient.mtu_size
"""

import asyncio

from bleak import BleakBackend, BleakClient, BleakScanner
from bleak.backends.scanner import AdvertisementData, BLEDevice

# replace with real characteristic UUID
CHAR_UUID = "00000000-0000-0000-0000-000000000000"


async def main() -> None:
    queue: asyncio.Queue[BLEDevice] = asyncio.Queue()

    def callback(device: BLEDevice, adv: AdvertisementData) -> None:
        # can use advertising data to filter here
        queue.put_nowait(device)

    async with BleakScanner(callback):
        # get the first matching device
        device = await queue.get()

    async with BleakClient(device) as client:
        # BlueZ doesn't have a proper way to get the MTU, so we have this hack.
        # If this doesn't work for you, you can set the client._mtu_size attribute
        # to override the value instead.
        if client.backend_id == BleakBackend.BLUEZ_DBUS:
            await client._backend._acquire_mtu()  # type: ignore

        print("MTU:", client.mtu_size)

        # Write without response is limited to MTU - 3 bytes

        data = bytes(1000)  # replace with real data
        chunk_size = client.mtu_size - 3
        for chunk in (
            data[i : i + chunk_size] for i in range(0, len(data), chunk_size)
        ):
            await client.write_gatt_char(CHAR_UUID, chunk, response=False)


if __name__ == "__main__":
    asyncio.run(main())



import sys
from typing import TYPE_CHECKING

import pytest
from bumble.device import Device

from bleak import BleakClient
from bleak.backends import BleakBackend, get_default_backend
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    find_ble_device,
)


@pytest.mark.skipif(
    get_default_backend() != BleakBackend.CORE_BLUETOOTH,
    reason="'get_rssi()' is only available on CoreBluetooth",
)
async def test_get_rssi(bumble_peripheral: Device):
    """Getting RSSI from client is possible."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device) as client:
        if TYPE_CHECKING:
            if sys.platform != "darwin":
                assert False, "This backend is only available on macOS"
        from bleak.backends.corebluetooth.client import BleakClientCoreBluetooth

        backend = client._backend  # pyright: ignore[reportPrivateUsage]
        assert isinstance(backend, BleakClientCoreBluetooth)
        rssi = await backend.get_rssi()

        # Verify that this value is an integer and not some other
        # type from a ffi binding framework.
        assert isinstance(rssi, int)

        # The rssi can vary. So we only check for a plausible range.
        assert -127 <= rssi < 0


async def test_mtu_size(bumble_peripheral: Device):
    """Check if the mtu size can be optained."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device) as client:
        if client.backend_id == BleakBackend.BLUEZ_DBUS:
            with pytest.warns(UserWarning):
                mtu_size = client.mtu_size
        else:
            mtu_size = client.mtu_size

        # Verify that this value is an integer and not some other
        # type from a ffi binding framework.
        assert isinstance(mtu_size, int)

        # The mtu_size is different between different platforms. So we only check
        # for a plausible range.
        assert 23 <= mtu_size <= 517



# Created on 2019-03-19 by hbldh <henrik.blidh@nedomkull.com>
"""
Interface class for the Bleak representation of a GATT Descriptor
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from bleak.uuids import normalize_uuid_16

# avoid circular import
if TYPE_CHECKING:
    from bleak.backends.characteristic import BleakGATTCharacteristic


_descriptor_descriptions = {
    normalize_uuid_16(0x2905): [
        "Characteristic Aggregate Format",
        "org.bluetooth.descriptor.gatt.characteristic_aggregate_format",
        "0x2905",
        "GSS",
    ],
    normalize_uuid_16(0x2900): [
        "Characteristic Extended Properties",
        "org.bluetooth.descriptor.gatt.characteristic_extended_properties",
        "0x2900",
        "GSS",
    ],
    normalize_uuid_16(0x2904): [
        "Characteristic Presentation Format",
        "org.bluetooth.descriptor.gatt.characteristic_presentation_format",
        "0x2904",
        "GSS",
    ],
    normalize_uuid_16(0x2901): [
        "Characteristic User Description",
        "org.bluetooth.descriptor.gatt.characteristic_user_description",
        "0x2901",
        "GSS",
    ],
    normalize_uuid_16(0x2902): [
        "Client Characteristic Configuration",
        "org.bluetooth.descriptor.gatt.client_characteristic_configuration",
        "0x2902",
        "GSS",
    ],
    normalize_uuid_16(0x290B): [
        "Environmental Sensing Configuration",
        "org.bluetooth.descriptor.es_configuration",
        "0x290B",
        "GSS",
    ],
    normalize_uuid_16(0x290C): [
        "Environmental Sensing Measurement",
        "org.bluetooth.descriptor.es_measurement",
        "0x290C",
        "GSS",
    ],
    normalize_uuid_16(0x290D): [
        "Environmental Sensing Trigger Setting",
        "org.bluetooth.descriptor.es_trigger_setting",
        "0x290D",
        "GSS",
    ],
    normalize_uuid_16(0x2907): [
        "External Report Reference",
        "org.bluetooth.descriptor.external_report_reference",
        "0x2907",
        "GSS",
    ],
    normalize_uuid_16(0x2909): [
        "Number of Digitals",
        "org.bluetooth.descriptor.number_of_digitals",
        "0x2909",
        "GSS",
    ],
    normalize_uuid_16(0x2908): [
        "Report Reference",
        "org.bluetooth.descriptor.report_reference",
        "0x2908",
        "GSS",
    ],
    normalize_uuid_16(0x2903): [
        "Server Characteristic Configuration",
        "org.bluetooth.descriptor.gatt.server_characteristic_configuration",
        "0x2903",
        "GSS",
    ],
    normalize_uuid_16(0x290E): [
        "Time Trigger Setting",
        "org.bluetooth.descriptor.time_trigger_setting",
        "0x290E",
        "GSS",
    ],
    normalize_uuid_16(0x2906): [
        "Valid Range",
        "org.bluetooth.descriptor.valid_range",
        "0x2906",
        "GSS",
    ],
    normalize_uuid_16(0x290A): [
        "Value Trigger Setting",
        "org.bluetooth.descriptor.value_trigger_setting",
        "0x290A",
        "GSS",
    ],
}


class BleakGATTDescriptor:
    """The Bleak representation of a GATT Descriptor"""

    def __init__(
        self, obj: Any, handle: int, uuid: str, characteristic: BleakGATTCharacteristic
    ):
        """
        Args:
            obj: The backend-specific object for the descriptor.
            handle: The handle of the descriptor.
            uuid: The UUID of the descriptor.
            characteristic: The characteristic that this descriptor belongs to.
        """
        self.obj = obj
        self._handle = handle
        self._uuid = uuid
        self._characteristic = characteristic

    def __str__(self):
        return f"{self.uuid} (Handle: {self.handle}): {self.description}"

    @property
    def characteristic_uuid(self) -> str:
        """UUID for the characteristic that this descriptor belongs to"""
        return self._characteristic.uuid

    @property
    def characteristic_handle(self) -> int:
        """handle for the characteristic that this descriptor belongs to"""
        return self._characteristic.handle

    @property
    def uuid(self) -> str:
        """UUID for this descriptor"""
        return self._uuid

    @property
    def handle(self) -> int:
        """Integer handle for this descriptor"""
        return self._handle

    @property
    def description(self) -> str:
        """A text description of what this descriptor represents"""
        return _descriptor_descriptions.get(self.uuid, ["Unknown"])[0]



# Created on 2017-11-19 by hbldh <henrik.blidh@nedomkull.com>
"""
__init__.py
"""
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

import objc

objc.options.verbose = True



#!/usr/bin/env python

"""Tests for `bleak.backends.bluezdbus.version` package."""

import sys

import pytest

if sys.platform != "linux":
    pytest.skip("skipping linux-only tests", allow_module_level=True)
    assert False  # HACK: work around pyright bug

from unittest.mock import AsyncMock, Mock

from bleak.backends.bluezdbus.version import BlueZFeatures


@pytest.fixture(autouse=True)
def reset_bluez_features(monkeypatch: pytest.MonkeyPatch) -> None:
    """Reset BlueZFeatures class variables to avoid side effects between tests."""
    monkeypatch.setattr(BlueZFeatures, "checked_bluez_version", False)
    monkeypatch.setattr(BlueZFeatures, "_check_bluez_event", None)
    monkeypatch.setattr(BlueZFeatures, "supported_version", True)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "version",
    [
        (b"bluetoothctl: 5.51"),
        (b"bluetoothctl: 5.63"),
        (b""),
    ],
)
async def test_bluez_version(
    version: bytes,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test we can determine supported feature from bluetoothctl."""
    mock_proc = Mock(
        wait=AsyncMock(), stdout=Mock(read=AsyncMock(return_value=version))
    )
    monkeypatch.setattr(
        "bleak.backends.bluezdbus.version.asyncio.create_subprocess_exec",
        AsyncMock(return_value=mock_proc),
    )
    await BlueZFeatures.check_bluez_version()
    assert BlueZFeatures.checked_bluez_version is True


@pytest.mark.asyncio
async def test_bluez_version_only_happens_once(monkeypatch: pytest.MonkeyPatch):
    """Test we can determine supported feature from bluetoothctl."""
    mock_proc = Mock(
        wait=AsyncMock(),
        stdout=Mock(read=AsyncMock(return_value=b"bluetoothctl: 5.46")),
    )
    monkeypatch.setattr(
        "bleak.backends.bluezdbus.version.asyncio.create_subprocess_exec",
        AsyncMock(return_value=mock_proc),
    )
    await BlueZFeatures.check_bluez_version()

    assert BlueZFeatures.checked_bluez_version is True

    monkeypatch.setattr(
        "bleak.backends.bluezdbus.version.asyncio.create_subprocess_exec",
        AsyncMock(side_effect=Exception),
    )
    await BlueZFeatures.check_bluez_version()

    assert BlueZFeatures.checked_bluez_version is True


@pytest.mark.asyncio
async def test_exception_checking_bluez_features_does_not_block_forever(
    monkeypatch: pytest.MonkeyPatch,
):
    """Test an exception while checking BlueZ features does not stall a second check."""
    monkeypatch.setattr(
        "bleak.backends.bluezdbus.version.asyncio.create_subprocess_exec",
        AsyncMock(side_effect=OSError),
    )
    await BlueZFeatures.check_bluez_version()

    assert BlueZFeatures.checked_bluez_version is True

    await BlueZFeatures.check_bluez_version()

    assert BlueZFeatures.checked_bluez_version is True



import asyncio
import logging
from collections.abc import Callable
from typing import Any

from bleak._compat import TypeVarTuple, Unpack

_Ts = TypeVarTuple("_Ts")


logger = logging.getLogger(__name__)


def try_call_soon_threadsafe(
    event_loop: asyncio.AbstractEventLoop,
    callback: Callable[[Unpack[_Ts]], Any],
    *args: Unpack[_Ts],
) -> None:
    """Call a callback on the event loop thread, handling closed loop errors gracefully."""
    try:
        event_loop.call_soon_threadsafe(callback, *args)
    except RuntimeError:
        # Likely caused by loop being closed
        logger.debug("unraisable exception", exc_info=True)



import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

from typing import Optional, overload

from CoreBluetooth import CBUUID
from Foundation import NSNumber, NSString

from bleak.uuids import normalize_uuid_str


def cb_uuid_to_str(uuid: CBUUID) -> str:
    """Converts a CoreBluetooth UUID to a Python string.

    If ``uuid`` is a 16-bit UUID, it is assumed to be a Bluetooth GATT UUID
    (``0000xxxx-0000-1000-8000-00805f9b34fb``).

    Args
        uuid: The UUID.

    Returns:
        The UUID as a lower case Python string (``xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxx``)
    """
    return normalize_uuid_str(uuid.UUIDString())


@overload
def to_optional_str(value: NSString) -> str: ...
@overload
def to_optional_str(value: None) -> None: ...


def to_optional_str(value: Optional[NSString]) -> Optional[str]:
    """Converts an NSString to a Python string or None.

    Args:
        value: The NSString or None.

    Returns:
        The Python string or None.
    """
    if value is None:
        return None

    return str(value)


@overload
def to_optional_int(value: NSNumber) -> int: ...
@overload
def to_optional_int(value: None) -> None: ...


def to_optional_int(value: Optional[NSNumber]) -> Optional[int]:
    """Converts an NSNumber to a Python int or None.

    Args:
        value: The NSNumber or None.

    Returns:
        The Python int or None.
    """
    if value is None:
        return None

    return int(value)



import sys

import pytest

# isort: off

if not sys.platform.startswith("darwin"):
    pytest.skip("backend only available on macOS", allow_module_level=True)


def test_deprecated_CBScannerArgs_import():
    with pytest.warns(
        DeprecationWarning,
        match="importing CBScannerArgs from bleak.backends.corebluetooth.scanner is deprecated, use bleak.args.corebluetooth instead",
    ) as recorder:
        from bleak.backends.corebluetooth.scanner import (  # noqa: F401
            CBScannerArgs,  # type: ignore[unused-import]
        )

    assert recorder.list[0].filename == __file__



# Created on 2019-06-26 by kevincar <kevincarrolldavis@gmail.com>
"""
BLE Client for CoreBluetooth on macOS
"""

import functools
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

import asyncio
import logging
from typing import Any, Optional, Union

from CoreBluetooth import (
    CBUUID,
    CBCharacteristicWriteWithoutResponse,
    CBCharacteristicWriteWithResponse,
    CBPeripheral,
    CBPeripheralStateConnected,
)
from Foundation import NSArray, NSData

from bleak import BleakScanner
from bleak._compat import override
from bleak.args import SizedBuffer
from bleak.args.corebluetooth import CBStartNotifyArgs
from bleak.assigned_numbers import gatt_char_props_to_strs
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.client import BaseBleakClient, NotifyCallback
from bleak.backends.corebluetooth.CentralManagerDelegate import CentralManagerDelegate
from bleak.backends.corebluetooth.PeripheralDelegate import PeripheralDelegate
from bleak.backends.corebluetooth.scanner import BleakScannerCoreBluetooth
from bleak.backends.corebluetooth.utils import cb_uuid_to_str
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
from bleak.exc import BleakDeviceNotFoundError, BleakError

logger = logging.getLogger(__name__)


class BleakClientCoreBluetooth(BaseBleakClient):
    """CoreBluetooth class interface for BleakClient

    Args:
        address_or_ble_device (`BLEDevice` or str): The Bluetooth address of the BLE peripheral to connect to or the `BLEDevice` object representing it.
        services: Optional set of service UUIDs that will be used.
    """

    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        services: Optional[set[str]] = None,
        **kwargs: Any,
    ):
        super().__init__(address_or_ble_device, **kwargs)

        self._peripheral: Optional[CBPeripheral] = None
        self._delegate: Optional[PeripheralDelegate] = None
        self._central_manager_delegate: Optional[CentralManagerDelegate] = None

        if isinstance(address_or_ble_device, BLEDevice):
            (
                self._peripheral,
                self._central_manager_delegate,
            ) = address_or_ble_device.details

        self._requested_services = (
            NSArray[CBUUID]
            .alloc()
            .initWithArray_(list(map(CBUUID.UUIDWithString_, services)))
            if services
            else None
        )

    def __str__(self) -> str:
        return f"BleakClientCoreBluetooth ({self.address})"

    @override
    async def connect(self, pair: bool, **kwargs: Any) -> None:
        """Connect to a specified Peripheral

        Keyword Args:
            timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call.
        """
        if pair:
            logger.debug("Explicit pairing is not available in CoreBluetooth.")

        timeout = kwargs.get("timeout", self._timeout)
        if self._peripheral is None:
            device = await BleakScanner.find_device_by_address(
                self.address, timeout=timeout, backend=BleakScannerCoreBluetooth
            )

            if device:
                self._peripheral, self._central_manager_delegate = device.details
            else:
                raise BleakDeviceNotFoundError(
                    self.address, f"Device with address {self.address} was not found"
                )

        if self._delegate is None:
            assert self._peripheral is not None
            self._delegate = PeripheralDelegate(self._peripheral)

        def disconnect_callback() -> None:
            # Ensure that `get_services` retrieves services again, rather
            # than using the cached object
            self.services = None

            assert self._delegate is not None

            # If there are any pending futures waiting for delegate callbacks, we
            # need to raise an exception since the callback will no longer be
            # called because the device is disconnected.
            for future in self._delegate.futures():
                try:
                    future.set_exception(BleakError("disconnected"))
                except asyncio.InvalidStateError:
                    # the future was already done
                    pass

            if self._disconnected_callback:
                self._disconnected_callback()

        manager = self._central_manager_delegate
        assert manager is not None
        logger.debug("CentralManagerDelegate at %r", manager)
        logger.debug("Connecting to BLE device @ %s", self.address)
        assert self._peripheral is not None
        await manager.connect(self._peripheral, disconnect_callback, timeout=timeout)

        # Now get services
        await self._get_services()

    @override
    async def disconnect(self) -> None:
        """Disconnect from the peripheral device"""
        if (
            self._peripheral is None
            or self._peripheral.state() != CBPeripheralStateConnected
        ):
            return

        assert self._central_manager_delegate
        await self._central_manager_delegate.disconnect(self._peripheral)

    @property
    @override
    def is_connected(self) -> bool:
        """Checks for current active connection"""
        return (
            False
            if self._peripheral is None
            else self._peripheral.state() == CBPeripheralStateConnected
        )

    @property
    @override
    def name(self) -> str:
        """Get the name of the connected peripheral"""
        if self._peripheral is None:
            raise BleakError("Not connected")
        return self._peripheral.name()

    @property
    @override
    def mtu_size(self) -> int:
        """Get ATT MTU size for active connection"""
        # Use type CBCharacteristicWriteWithoutResponse to get maximum write
        # value length based on the negotiated ATT MTU size. Add the ATT header
        # length (+3) to get the actual ATT MTU size.
        assert self._peripheral
        return (
            self._peripheral.maximumWriteValueLengthForType_(
                CBCharacteristicWriteWithoutResponse
            )
            + 3
        )

    @override
    async def pair(self, *args: Any, **kwargs: Any) -> None:
        """Attempt to pair with a peripheral.

        Raises:
            NotImplementedError:
                This is not available on macOS since there is not explicit API
                to do a pairing. Instead, the docs state that it "auto-pairs",
                when trying to read a characteristic that requires encryption.

        Reference:

            - `Apple Docs <https://developer.apple.com/library/archive/documentation/NetworkingInternetWeb/Conceptual/CoreBluetooth_concepts/BestPracticesForSettingUpYourIOSDeviceAsAPeripheral/BestPracticesForSettingUpYourIOSDeviceAsAPeripheral.html#//apple_ref/doc/uid/TP40013257-CH5-SW1>`_
            - `Stack Overflow post #1 <https://stackoverflow.com/questions/25254932/can-you-pair-a-bluetooth-le-device-in-an-ios-app>`_
            - `Stack Overflow post #2 <https://stackoverflow.com/questions/47546690/ios-bluetooth-pairing-request-dialog-can-i-know-the-users-choice>`_
        """
        raise NotImplementedError("Pairing is not available in Core Bluetooth.")

    @override
    async def unpair(self) -> None:
        """
        Remove pairing information for a peripheral.

        Raises:
            NotImplementedError:
                This is not available on macOS since there is not explicit API
                to do a pairing.
        """
        raise NotImplementedError("Pairing is not available in Core Bluetooth.")

    async def _get_services(self) -> BleakGATTServiceCollection:
        """Get all services registered for this GATT server.

        Returns:
           A :py:class:`bleak.backends.service.BleakGATTServiceCollection` with this device's services tree.

        """
        if self.services is not None:
            return self.services

        services = BleakGATTServiceCollection()

        logger.debug("Retrieving services...")
        assert self._delegate
        assert self._peripheral
        cb_services = await self._delegate.discover_services(self._requested_services)

        for service in cb_services:
            serv = BleakGATTService(
                service, service.startHandle(), cb_uuid_to_str(service.UUID())
            )
            services.add_service(serv)

            serviceUUID = service.UUID().UUIDString()
            logger.debug("Retrieving characteristics for service %s", serviceUUID)
            characteristics = await self._delegate.discover_characteristics(service)

            for characteristic in characteristics:
                cUUID = characteristic.UUID().UUIDString()
                logger.debug("Retrieving descriptors for characteristic %s", cUUID)

                char = BleakGATTCharacteristic(
                    characteristic,
                    characteristic.handle(),
                    cb_uuid_to_str(characteristic.UUID()),
                    list(gatt_char_props_to_strs(characteristic.properties())),
                    functools.partial(
                        self._peripheral.maximumWriteValueLengthForType_,
                        CBCharacteristicWriteWithoutResponse,
                    ),
                    serv,
                )
                services.add_characteristic(char)

                descriptors = await self._delegate.discover_descriptors(characteristic)
                for descriptor in descriptors:
                    desc = BleakGATTDescriptor(
                        descriptor,
                        int(descriptor.handle()),
                        cb_uuid_to_str(descriptor.UUID()),
                        char,
                    )
                    services.add_descriptor(desc)

        logger.debug("Services resolved for %s", str(self))
        self.services = services
        return self.services

    @override
    async def read_gatt_char(
        self, characteristic: BleakGATTCharacteristic, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to read from.

        Returns:
            (bytearray) The read data.

        """
        assert self._delegate
        output = await self._delegate.read_characteristic(
            characteristic.obj, use_cached=kwargs.get("use_cached", False)
        )
        value = bytearray(output)
        logger.debug("Read Characteristic %s: %r", characteristic.uuid, value)
        return value

    @override
    async def read_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT descriptor.

        Args:
            handle (int): The handle of the descriptor to read from.
            use_cached (bool): `False` forces Windows to read the value from the
                device again and not use its own cached value. Defaults to `False`.

        Returns:
            (bytearray) The read data.
        """
        assert self._delegate
        output = await self._delegate.read_descriptor(
            descriptor.obj, use_cached=kwargs.get("use_cached", False)
        )
        if isinstance(
            output, str
        ):  # Sometimes a `pyobjc_unicode`or `__NSCFString` is returned and they can be used as regular Python strings.
            value = bytearray(output.encode("utf-8"))
        else:  # _NSInlineData
            value = bytearray(output)  # value.getBytes_length_(None, len(value))
        logger.debug("Read Descriptor %d : %r", descriptor.handle, value)
        return value

    @override
    async def write_gatt_char(
        self, characteristic: BleakGATTCharacteristic, data: SizedBuffer, response: bool
    ) -> None:
        value = NSData.alloc().initWithBytes_length_(data, len(data))
        assert self._delegate
        await self._delegate.write_characteristic(
            characteristic.obj,
            value,
            (
                CBCharacteristicWriteWithResponse
                if response
                else CBCharacteristicWriteWithoutResponse
            ),
        )
        logger.debug(f"Write Characteristic {characteristic.uuid} : {data}")

    @override
    async def write_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, data: SizedBuffer
    ) -> None:
        """Perform a write operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.
            data: The data to send (any bytes-like object).

        """
        assert self._delegate
        value = NSData.alloc().initWithBytes_length_(data, len(data))
        await self._delegate.write_descriptor(descriptor.obj, value)
        logger.debug("Write Descriptor %d : %r", descriptor.handle, data)

    @override
    async def start_notify(
        self,
        characteristic: BleakGATTCharacteristic,
        callback: NotifyCallback,
        **kwargs: Any,
    ) -> None:
        """
        Activate notifications/indications on a characteristic.
        """
        assert self._delegate is not None

        cb: CBStartNotifyArgs = kwargs["cb"]

        await self._delegate.start_notifications(
            characteristic.obj,
            callback,
            cb.get("notification_discriminator"),
        )

    @override
    async def stop_notify(self, characteristic: BleakGATTCharacteristic) -> None:
        """Deactivate notification/indication on a specified characteristic.

        Args:
            characteristic (BleakGATTCharacteristic: The characteristic to deactivate
                notification/indication on.
        """
        assert self._delegate
        await self._delegate.stop_notifications(characteristic.obj)

    async def get_rssi(self) -> int:
        """To get RSSI value in dBm of the connected Peripheral"""
        assert self._delegate
        return int(await self._delegate.read_rssi())



import asyncio

import pytest
from bumble.device import Connection, Device
from bumble.gatt import Characteristic, CharacteristicValue, Service

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    find_ble_device,
)


async def test_read_gatt_char(bumble_peripheral: Device):
    """Reading a GATT characteristic is possible."""
    READ_SERVICE_UUID = "2908f536-3fab-43c9-a7b2-80b6fdaae99b"
    READ_CHARACTERISITC_UUID = "1a1af049-9c23-4b69-b763-e1096674ed18"
    virtual_characteristic = Characteristic(
        READ_CHARACTERISITC_UUID,
        Characteristic.Properties.READ,
        Characteristic.Permissions.READABLE,
        b"DATA",
    )
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[Service(READ_SERVICE_UUID, [virtual_characteristic])],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device, services=[READ_SERVICE_UUID]) as client:
        data = await client.read_gatt_char(READ_CHARACTERISITC_UUID)
        assert data == b"DATA"


async def test_write_gatt_char_with_response(bumble_peripheral: Device):
    """Writing a GATT characteristic is possible."""
    WRITE_WITH_RESPONSE_SERVICE_UUID = "79a92bad-31b4-4a70-885c-e704ae2c6363"
    WRITE_WITH_RESPONSE_CHARACTERISITC_UUID = "2a07f6ea-6401-45d7-838c-0f459a7edb7f"
    virtual_characteristic = Characteristic(
        WRITE_WITH_RESPONSE_CHARACTERISITC_UUID,
        Characteristic.Properties.WRITE,
        Characteristic.WRITEABLE,
        b"----",
    )
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[Service(WRITE_WITH_RESPONSE_SERVICE_UUID, [virtual_characteristic])],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(
        device, services=[WRITE_WITH_RESPONSE_SERVICE_UUID]
    ) as client:
        await client.write_gatt_char(
            WRITE_WITH_RESPONSE_CHARACTERISITC_UUID, b"DATA", response=True
        )
        assert virtual_characteristic.value == b"DATA"


async def test_write_gatt_char_no_response(bumble_peripheral: Device):
    """Writing a GATT characteristic is possible."""
    WRITE_WITHOUT_RESPONSE_SERVICE_UUID = "79a92bad-31b4-4a70-885c-e704ae2c6363"
    WRITE_WITHOUT_RESPONSE_CHARACTERISITC_UUID = "2a07f6ea-6401-45d7-838c-0f459a7edb7f"

    peripheral_write_callback_called: asyncio.Future[bytes] = asyncio.Future()

    def peripheral_write_callback(connection: Connection, value: bytes):
        peripheral_write_callback_called.set_result(value)

    virtual_characteristic = Characteristic[bytes](
        WRITE_WITHOUT_RESPONSE_CHARACTERISITC_UUID,
        Characteristic.Properties.WRITE_WITHOUT_RESPONSE,
        Characteristic.WRITEABLE,
        CharacteristicValue(write=peripheral_write_callback),
    )
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[
            Service(WRITE_WITHOUT_RESPONSE_SERVICE_UUID, [virtual_characteristic])
        ],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(
        device, services=[WRITE_WITHOUT_RESPONSE_SERVICE_UUID]
    ) as client:
        await client.write_gatt_char(
            WRITE_WITHOUT_RESPONSE_CHARACTERISITC_UUID, b"DATA", response=False
        )
        written_value = await asyncio.wait_for(
            peripheral_write_callback_called, timeout=1
        )
        assert written_value == b"DATA"


async def test_notify_gatt_char(bumble_peripheral: Device):
    """Writing a GATT characteristic is possible."""
    NOTIFY_SERVICE_UUID = "e405a09d-7c8e-4ac5-adcf-ba808e7f2d18"
    NOTIFY_CHARACTERISITC_UUID = "d4c6dad3-76f1-4034-8871-0a6345be6cfc"
    virtual_characteristic = Characteristic(
        NOTIFY_CHARACTERISITC_UUID,
        Characteristic.Properties.READ | Characteristic.Properties.NOTIFY,
        Characteristic.Permissions.READABLE,
        b"----",
    )
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[Service(NOTIFY_SERVICE_UUID, [virtual_characteristic])],
    )

    device = await find_ble_device(bumble_peripheral)

    notified_data: asyncio.Queue[bytes] = asyncio.Queue()

    def notify_callback(characteristic: BleakGATTCharacteristic, data: bytearray):
        assert characteristic.uuid.lower() == NOTIFY_CHARACTERISITC_UUID
        notified_data.put_nowait(bytes(data))

    async with BleakClient(device, services=[NOTIFY_SERVICE_UUID]) as client:
        await client.start_notify(
            NOTIFY_CHARACTERISITC_UUID,
            notify_callback,
        )
        assert notified_data.empty()

        await bumble_peripheral.notify_subscribers(  # type: ignore  # (missing type hints in bumble)
            virtual_characteristic,
            b"1234",
        )

        data = await asyncio.wait_for(notified_data.get(), timeout=1)
        assert data == b"1234"

        await bumble_peripheral.notify_subscribers(  # type: ignore  # (missing type hints in bumble)
            virtual_characteristic,
            b"2345",
        )

        data = await asyncio.wait_for(notified_data.get(), timeout=1)
        assert data == b"2345"

        await client.stop_notify(NOTIFY_CHARACTERISITC_UUID)

        await bumble_peripheral.notify_subscribers(  # type: ignore  # (missing type hints in bumble)
            virtual_characteristic,
            b"2345",
        )

        # Verify no notification was received
        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(notified_data.get(), timeout=1)



from bumble.device import Device
from bumble.gatt import (
    GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
    Characteristic,
    Descriptor,
    Service,
)

from bleak import BleakClient
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    find_ble_device,
)


async def test_read_gatt_descriptor(bumble_peripheral: Device):
    """Reading a GATT descriptor is possible."""
    READABLE_DESCRIPTOR_SERVICE_UUID = "0d15eded-4e68-4718-bedf-736847b68e72"
    READABLE_DESCRIPTOR_CHARACTERISITC_UUID = "25c614ab-1560-46da-94bc-c146addfc359"
    virtual_descriptor = Descriptor(
        GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
        Descriptor.READABLE,
        "Description".encode(),
    )
    virtual_characteristic = Characteristic(
        READABLE_DESCRIPTOR_CHARACTERISITC_UUID,
        Characteristic.Properties.READ,
        Characteristic.Permissions.READABLE,
        b"",
        [virtual_descriptor],
    )
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[Service(READABLE_DESCRIPTOR_SERVICE_UUID, [virtual_characteristic])],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(
        device, services=[READABLE_DESCRIPTOR_SERVICE_UUID]
    ) as client:
        characteristic = client.services.get_characteristic(
            READABLE_DESCRIPTOR_CHARACTERISITC_UUID
        )
        assert characteristic

        descriptor = characteristic.get_descriptor(
            GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR.to_hex_str()
        )
        assert descriptor

        data = await client.read_gatt_descriptor(descriptor)
        assert data == b"Description"


async def test_write_gatt_descriptor(bumble_peripheral: Device):
    """Writing a GATT descriptor is possible."""
    WRITABLE_DESCRIPTOR_SERVICE_UUID = "bef6dc41-8986-4c32-b746-6e2b10ca06e0"
    WRITABLE_DESCRIPTOR_CHARACTERISITC_UUID = "822afd2f-c2b2-4302-9edb-09850a93b707"
    virtual_descriptor = Descriptor(
        GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
        Descriptor.WRITEABLE,
        b"-----------",
    )
    virtual_characteristic = Characteristic(
        WRITABLE_DESCRIPTOR_CHARACTERISITC_UUID,
        Characteristic.Properties.READ,
        Characteristic.Permissions.READABLE,
        b"",
        [virtual_descriptor],
    )
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[Service(WRITABLE_DESCRIPTOR_SERVICE_UUID, [virtual_characteristic])],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(
        device, services=[WRITABLE_DESCRIPTOR_SERVICE_UUID]
    ) as client:
        characteristic = client.services.get_characteristic(
            WRITABLE_DESCRIPTOR_CHARACTERISITC_UUID
        )
        assert characteristic

        descriptor = characteristic.get_descriptor(
            GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR.to_hex_str()
        )
        assert descriptor

        await client.write_gatt_descriptor(descriptor, b"Description")
        assert virtual_descriptor.value == b"Description"  # type: ignore  # (missing type hints in bumble)



import asyncio

from bumble.att import Attribute, AttributeValue
from bumble.device import Connection, Device
from bumble.gatt import (
    GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
    Characteristic,
    Descriptor,
    Service,
)

from bleak import BleakClient, BleakScanner
from bleak.backends.characteristic import BleakGATTCharacteristic
from tests.integration.conftest import add_default_advertising_data

TEST_SERVICE_UUID = "9d513f40-5c89-42dc-9688-2cfa30f2d9e7"
TEST_CHARACTERISTIC_UUID = "e809cb2f-34e3-42a1-ba92-22db2495cd6a"


async def test_notification_sent_before_write_response(
    bumble_peripheral: Device,
) -> None:
    """
    Regression test for <https://github.com/hbldh/bleak/issues/1885>.
    """

    notifications_enabled = False

    def on_cccd_read(connection: Connection) -> bytes:
        return b"\x01\x00" if notifications_enabled else b"\x00\x00"

    async def on_cccd_write(connection: Connection, value: bytes) -> None:
        nonlocal notifications_enabled
        notifications_enabled = value == b"\x01\x00"

        # This is simulating an unusual peripheral that sends a notification
        # immediately upon receiving the CCCD write, before sending the write
        # response.

        # TODO: Type hints in bumble need to be fixed to be able to remove the pyright ignore
        await bumble_peripheral.notify_subscribers(  # pyright: ignore[reportUnknownMemberType]
            test_characteristic, b"test", force=True
        )

    cccd_value = AttributeValue[bytes](on_cccd_read, on_cccd_write)

    test_characteristic = Characteristic[bytes](
        TEST_CHARACTERISTIC_UUID,
        Characteristic.Properties.NOTIFY,
        Attribute.Permissions(0),
        descriptors=[
            Descriptor(
                GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
                Attribute.Permissions.WRITEABLE | Attribute.Permissions.READABLE,
                cccd_value,
            )
        ],
    )

    bumble_peripheral.add_service(Service(TEST_SERVICE_UUID, [test_characteristic]))

    add_default_advertising_data(bumble_peripheral)
    await bumble_peripheral.power_on()
    await bumble_peripheral.start_advertising()

    device = await BleakScanner.find_device_by_address(
        bumble_peripheral.static_address.to_string(), cb={"use_bdaddr": True}
    )

    assert device is not None, "Could not find bumble peripheral device"

    async with BleakClient(device, services=[TEST_SERVICE_UUID]) as client:
        notification_queue: asyncio.Queue[bytes] = asyncio.Queue()

        def on_notification(_: BleakGATTCharacteristic, data: bytearray) -> None:
            notification_queue.put_nowait(bytes(data))

        await client.start_notify(
            TEST_CHARACTERISTIC_UUID, on_notification, bluez={"use_start_notify": True}
        )

        # In BlueZ, the notification is not received when using "AcquireNotify"
        # causing this to timeout.

        data = await asyncio.wait_for(notification_queue.get(), timeout=3)

        assert data == b"test"



import sys
from collections.abc import Iterator, Mapping, Sequence
from typing import Any, NewType, Optional, SupportsIndex, TypeVar, overload

if sys.version_info < (3, 12):
    from typing_extensions import Buffer
else:
    from collections.abc import Buffer

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

TNSObject = TypeVar("TNSObject", bound=NSObject)

class NSObject:
    @classmethod
    def alloc(cls) -> Self: ...
    def init(self) -> Optional[Self]: ...
    def addObserver_forKeyPath_options_context_(
        self,
        observer: NSObject,
        keyPath: str,
        options: NSKeyValueObservingOptions,
        context: int,
    ) -> None: ...
    def removeObserver_forKeyPath_(self, observer: NSObject, keyPath: str) -> None: ...
    @classmethod
    def __init_subclass__(cls, /, protocols: Sequence[Any] = []) -> None: ...

_KT = TypeVar("_KT")
_VT_co = TypeVar("_VT_co", covariant=True)

class NSDictionary(NSObject, Mapping[_KT, _VT_co]):
    def __getitem__(self, key: _KT) -> _VT_co: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def __len__(self) -> int: ...

class NSUUID(NSObject):
    @classmethod
    def UUIDWithString_(cls, uuidString: str) -> NSUUID: ...
    def UUIDString(self) -> NSString: ...
    def isEqualToUUID_(self, other: NSUUID) -> bool: ...

class NSError(NSObject): ...

class NSData(NSObject, Buffer):
    def initWithBytes_length_(self, bytes: Buffer, length: int) -> Self: ...
    def length(self) -> int: ...
    def getBytes_length_(self, buffer: bytes, length: int) -> None: ...
    @overload
    def __getitem__(self, index: SupportsIndex) -> int: ...
    @overload
    def __getitem__(self, index: slice) -> memoryview: ...

T = TypeVar("T")

class NSArray(NSObject, Sequence[T]):
    @overload
    def __getitem__(self, index: int) -> T: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[T]: ...
    def __len__(self) -> int: ...
    def initWithArray_(self, array: Sequence[Any]) -> Self: ...

class NSValue(NSObject): ...
class NSString(NSObject, str): ...

class NSNumber(NSObject):
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __bool__(self) -> bool: ...

class NSBundle(NSObject):
    @classmethod
    def mainBundle(cls) -> NSBundle: ...
    def bundleIdentifier(self) -> NSString: ...

NSKeyValueObservingOptions = NewType("NSKeyValueObservingOptions", int)
NSKeyValueObservingOptionNew: NSKeyValueObservingOptions
NSKeyValueObservingOptionOld: NSKeyValueObservingOptions
NSKeyValueObservingOptionInitial: NSKeyValueObservingOptions
NSKeyValueObservingOptionPrior: NSKeyValueObservingOptions

NSKeyValueChangeKey = NewType("NSKeyValueChangeKey", str)
NSKeyValueChangeIndexesKey: NSKeyValueChangeKey
NSKeyValueChangeKindKey: NSKeyValueChangeKey
NSKeyValueChangeNewKey: NSKeyValueChangeKey
NSKeyValueChangeNotificationIsPriorKey: NSKeyValueChangeKey
NSKeyValueChangeOldKey: NSKeyValueChangeKey



import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "linux":
        assert False, "This backend is only available on Linux"

import asyncio
import contextlib
import logging
import re
from typing import Optional

logger = logging.getLogger(__name__)


async def _get_bluetoothctl_version() -> Optional[re.Match[bytes]]:
    """Get the version of bluetoothctl."""
    with contextlib.suppress(Exception):
        proc = await asyncio.create_subprocess_exec(
            "bluetoothctl", "--version", stdout=asyncio.subprocess.PIPE
        )
        assert proc.stdout
        out = await proc.stdout.read()
        version = re.search(b"(\\d+).(\\d+)", out.strip(b"'"))
        await proc.wait()
        return version
    return None


class BlueZFeatures:
    """Check which features are supported by the BlueZ backend."""

    checked_bluez_version = False
    supported_version = True
    _check_bluez_event: Optional[asyncio.Event] = None

    @classmethod
    async def check_bluez_version(cls) -> None:
        """Check the bluez version."""
        if cls._check_bluez_event:
            # If there is already a check in progress
            # it wins, wait for it instead
            await cls._check_bluez_event.wait()
            return
        cls._check_bluez_event = asyncio.Event()
        version_output = await _get_bluetoothctl_version()
        if version_output:
            major, minor = tuple(map(int, version_output.groups()))
            cls.supported_version = major == 5 and minor >= 55
        else:
            # Its possible they may be running inside a container where
            # bluetoothctl is not available and they only have access to the
            # BlueZ D-Bus API.
            logger.warning(
                "Could not determine BlueZ version, bluetoothctl not available, assuming 5.55+"
            )

        cls._check_bluez_event.set()
        cls.checked_bluez_version = True



from typing import Literal, Optional, TypeVar, overload

from CoreBluetooth import CBCentralManagerDelegate, CBPeripheralDelegate
from Foundation import NSObject

T = TypeVar("T")

class _OptionsType:
    deprecation_warnings: bool
    structs_indexable: bool
    structs_writable: bool
    unknown_pointer_raises: bool
    use_kvo: bool
    verbose: bool

options: _OptionsType

def super(cls: type, self: T) -> T: ...
def macos_available(major: int, minor: int, patch: int = 0) -> bool: ...
def python_method(func: T) -> T: ...

class WeakRef:
    def __init__(self, object: NSObject) -> None: ...
    def __call__(self) -> Optional[NSObject]: ...

@overload
def protocolNamed(
    name: Literal["CBCentralManagerDelegate"],
) -> type[CBCentralManagerDelegate]: ...
@overload
def protocolNamed(
    name: Literal["CBPeripheralDelegate"],
) -> type[CBPeripheralDelegate]: ...



#!/usr/bin/env python

"""Tests for `bleak.backends.bluezdbus.utils` package."""

import sys

import pytest

if sys.platform != "linux":
    pytest.skip("skipping linux-only tests", allow_module_level=True)
    assert False  # HACK: work around pyright bug


def test_device_path_from_characteristic_path():
    """Test device_path_from_characteristic_path."""
    from bleak.backends.bluezdbus.utils import (  # pylint: disable=import-outside-toplevel
        device_path_from_characteristic_path,
    )

    assert (
        device_path_from_characteristic_path(
            "/org/bluez/hci0/dev_11_22_33_44_55_66/service000c/char000d"
        )
        == "/org/bluez/hci0/dev_11_22_33_44_55_66"
    )

    assert (
        device_path_from_characteristic_path(
            "/org/bluez/hci10/dev_11_22_33_44_55_66/service000c/char000d"
        )
        == "/org/bluez/hci10/dev_11_22_33_44_55_66"
    )



"""
UART Service
-------------

An example showing how to write a simple program using the Nordic Semiconductor
(nRF) UART service.

"""

import asyncio
import sys
from itertools import count, takewhile
from typing import Iterator

from bleak import BleakClient, BleakScanner
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

UART_SERVICE_UUID = "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"
UART_RX_CHAR_UUID = "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"
UART_TX_CHAR_UUID = "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"


# TIP: you can get this function and more from the ``more-itertools`` package.
def sliced(data: bytes, n: int) -> Iterator[bytes]:
    """
    Slices *data* into chunks of size *n*. The last slice may be smaller than
    *n*.
    """
    return takewhile(len, (data[i : i + n] for i in count(0, n)))


async def uart_terminal():
    """This is a simple "terminal" program that uses the Nordic Semiconductor
    (nRF) UART service. It reads from stdin and sends each line of data to the
    remote device. Any data received from the device is printed to stdout.
    """

    def match_nus_uuid(device: BLEDevice, adv: AdvertisementData):
        # This assumes that the device includes the UART service UUID in the
        # advertising data. This test may need to be adjusted depending on the
        # actual advertising data supplied by the device.
        if UART_SERVICE_UUID.lower() in adv.service_uuids:
            return True

        return False

    device = await BleakScanner.find_device_by_filter(match_nus_uuid)

    if device is None:
        print("no matching device found, you may need to edit match_nus_uuid().")
        sys.exit(1)

    def handle_disconnect(_: BleakClient):
        print("Device was disconnected, goodbye.")
        # cancelling all tasks effectively ends the program
        for task in asyncio.all_tasks():
            task.cancel()

    def handle_rx(_: BleakGATTCharacteristic, data: bytearray):
        print("received:", data)

    async with BleakClient(device, disconnected_callback=handle_disconnect) as client:
        await client.start_notify(UART_TX_CHAR_UUID, handle_rx)

        print("Connected, start typing and press ENTER...")

        loop = asyncio.get_running_loop()
        nus = client.services.get_service(UART_SERVICE_UUID)
        assert nus is not None, "UART service not found"
        rx_char = nus.get_characteristic(UART_RX_CHAR_UUID)
        assert rx_char is not None, "UART RX characteristic not found"

        while True:
            # This waits until you type a line and press ENTER.
            # A real terminal program might put stdin in raw mode so that things
            # like CTRL+C get passed to the remote device.
            data = await loop.run_in_executor(None, sys.stdin.buffer.readline)

            # data will be empty on EOF (e.g. CTRL+D on *nix)
            if not data:
                break

            # some devices, like devices running MicroPython, expect Windows
            # line endings (uncomment line below if needed)
            # data = data.replace(b"\n", b"\r\n")

            # Writing without response requires that the data can fit in a
            # single BLE packet. We can use the max_write_without_response_size
            # property to split the data into chunks that will fit.

            for s in sliced(data, rx_char.max_write_without_response_size):
                await client.write_gatt_char(rx_char, s, response=False)

            print("sent:", data)


if __name__ == "__main__":
    try:
        asyncio.run(uart_terminal())
    except asyncio.CancelledError:
        # task is cancelled on disconnect, so we ignore this error
        pass



from setuptools import find_packages, setup

VERSION = "[VERSION]"  # Version will be filled in by the bleak recipe
NAME = "bleak"

setup(
    name=NAME,
    version=VERSION,
    packages=find_packages(exclude=("tests", "examples", "docs")),
)



"""
-------------------------------
CoreBluetooth backend arguments
-------------------------------
"""

from collections.abc import Callable
from typing import Optional, TypedDict


class CBScannerArgs(TypedDict, total=False):
    """
    Platform-specific :class:`BleakScanner` args for the CoreBluetooth backend.
    """

    use_bdaddr: bool
    """
    If true, use Bluetooth address instead of UUID.

    .. warning:: This uses an undocumented IOBluetooth API to get the Bluetooth
        address and may break in the future macOS releases. `It is known to not
        work on macOS 10.15 <https://github.com/hbldh/bleak/issues/1286>`_.
    """


NotificationDiscriminator = Callable[[bytes], bool]


class CBStartNotifyArgs(TypedDict, total=False):
    """CoreBluetooth backend-specific dictionary of arguments for the
    :meth:`bleak.BleakClient.start_notify` method.
    """

    notification_discriminator: Optional[NotificationDiscriminator]
    """
    A function that takes a single argument of a characteristic value
    and returns ``True`` if the value is from a notification or
    ``False`` if the value is from a read response.

    .. seealso:: :ref:`cb-notification-discriminator` for more info.
    """



"""
Scan/Discovery
--------------

Example showing how to scan for BLE devices.

Updated on 2019-03-25 by hbldh <henrik.blidh@nedomkull.com>

"""

import argparse
import asyncio

from bleak import BleakScanner


class Args(argparse.Namespace):
    macos_use_bdaddr: bool
    services: list[str]


async def main(args: Args):
    print("scanning for 5 seconds, please wait...")

    devices = await BleakScanner.discover(
        return_adv=True,
        service_uuids=args.services,
        cb={"use_bdaddr": args.macos_use_bdaddr},
    )

    for d, a in devices.values():
        print()
        print(d)
        print("-" * len(str(d)))
        print(a)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--services",
        metavar="<uuid>",
        nargs="*",
        help="UUIDs of one or more services to filter for",
    )

    parser.add_argument(
        "--macos-use-bdaddr",
        action="store_true",
        help="when true use Bluetooth address instead of UUID on macOS",
    )

    args = parser.parse_args(namespace=Args())

    asyncio.run(main(args))



import argparse
import asyncio
import contextlib
import logging
from typing import Iterable

from bleak import BleakClient, BleakScanner
from bleak.backends.characteristic import BleakGATTCharacteristic

logger = logging.getLogger(__name__)


class Args(argparse.Namespace):
    device1: str
    uuid1: str
    device2: str
    uuid2: str
    by_address: bool
    macos_use_bdaddr: bool
    debug: bool


async def connect_to_device(
    lock: asyncio.Lock,
    by_address: bool,
    macos_use_bdaddr: bool,
    name_or_address: str,
    notify_uuid: str,
):
    """
    Scan and connect to a device then print notifications for 10 seconds before
    disconnecting.

    Args:
        lock:
            The same lock must be passed to all calls to this function.
        by_address:
            If true, treat *name_or_address* as an address, otherwise treat
            it as a name.
        macos_use_bdaddr:
            If true, enable hack to allow use of Bluetooth address instead of
            UUID on macOS.
        name_or_address:
            The Bluetooth address/UUID of the device to connect to.
        notify_uuid:
            The UUID of a characteristic that supports notifications.
    """
    logger.info("starting %s task", name_or_address)

    try:
        async with contextlib.AsyncExitStack() as stack:

            # Trying to establish a connection to two devices at the same time
            # can cause errors, so use a lock to avoid this.
            async with lock:
                logger.info("scanning for %s", name_or_address)

                if by_address:
                    device = await BleakScanner.find_device_by_address(
                        name_or_address, cb={"use_bdaddr": macos_use_bdaddr}
                    )
                else:
                    device = await BleakScanner.find_device_by_name(name_or_address)

                logger.info("stopped scanning for %s", name_or_address)

                if device is None:
                    logger.error("%s not found", name_or_address)
                    return

                logger.info("connecting to %s", name_or_address)

                client = await stack.enter_async_context(BleakClient(device))

                logger.info("connected to %s", name_or_address)
                # This will be called immediately before client.__aexit__ when
                # the stack context manager exits.
                stack.callback(logger.info, "disconnecting from %s", name_or_address)

            # The lock is released here. The device is still connected and the
            # Bluetooth adapter is now free to scan and connect another device
            # without disconnecting this one.

            def callback(_: BleakGATTCharacteristic, data: bytearray) -> None:
                logger.info("%s received %r", name_or_address, data)

            await client.start_notify(notify_uuid, callback)
            await asyncio.sleep(10.0)
            await client.stop_notify(notify_uuid)

        # The stack context manager exits here, triggering disconnection.

        logger.info("disconnected from %s", name_or_address)

    except Exception:
        logger.exception("error with %s", name_or_address)


async def main(
    by_address: bool,
    macos_use_bdaddr: bool,
    addresses: Iterable[str],
    uuids: Iterable[str],
):
    lock = asyncio.Lock()

    await asyncio.gather(
        *(
            connect_to_device(lock, by_address, macos_use_bdaddr, address, uuid)
            for address, uuid in zip(addresses, uuids)
        )
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "device1",
        metavar="<device>",
        help="Bluetooth name or address of first device connect to",
    )
    parser.add_argument(
        "uuid1",
        metavar="<uuid>",
        help="notification characteristic UUID on first device",
    )
    parser.add_argument(
        "device2",
        metavar="<device>",
        help="Bluetooth name or address of second device to connect to",
    )
    parser.add_argument(
        "uuid2",
        metavar="<uuid>",
        help="notification characteristic UUID on second device",
    )

    parser.add_argument(
        "--by-address",
        action="store_true",
        help="when true treat <device> args as Bluetooth address instead of name",
    )

    parser.add_argument(
        "--macos-use-bdaddr",
        action="store_true",
        help="when true use Bluetooth address instead of UUID on macOS",
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="sets the log level to debug",
    )

    args = parser.parse_args(namespace=Args())

    log_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(
        level=log_level,
        format="%(asctime)-15s %(name)-8s %(levelname)s: %(message)s",
    )

    asyncio.run(
        main(
            args.by_address,
            args.macos_use_bdaddr,
            (args.device1, args.device2),
            (args.uuid1, args.uuid2),
        )
    )



#!/usr/bin/env python

"""Tests for `bleak.backends.winrt.util` package."""

import sys

import pytest

if sys.platform != "win32":
    pytest.skip("skipping windows-only tests", allow_module_level=True)
    # HACK: work around pyright bug
    allow_sta: object

from ctypes import windll, wintypes  # type: ignore[attr-defined]

from bleak.backends.winrt.util import (
    allow_sta,
    assert_mta,
    check_hresult,
    uninitialize_sta,
)
from bleak.exc import BleakError

# https://learn.microsoft.com/en-us/windows/win32/api/objbase/ne-objbase-coinit
COINIT_MULTITHREADED = 0x0
COINIT_APARTMENTTHREADED = 0x2

# https://learn.microsoft.com/en-us/windows/win32/api/combaseapi/nf-combaseapi-coinitializeex
_CoInitializeEx = windll.ole32.CoInitializeEx
_CoInitializeEx.restype = wintypes.LONG
_CoInitializeEx.argtypes = [wintypes.LPVOID, wintypes.DWORD]
_CoInitializeEx.errcheck = check_hresult  # type: ignore[assignment]

# https://learn.microsoft.com/en-us/windows/win32/api/combaseapi/nf-combaseapi-couninitialize
_CoUninitialize = windll.ole32.CoUninitialize
_CoUninitialize.restype = None
_CoUninitialize.argtypes = []


@pytest.fixture(autouse=True)
def run_around_tests():
    # reset global state
    try:
        delattr(allow_sta, "_allowed")
    except AttributeError:
        pass

    yield


@pytest.mark.asyncio
async def test_assert_mta_no_init():
    """Test device_path_from_characteristic_path."""

    await assert_mta()


@pytest.mark.asyncio
async def test_assert_mta_init_mta():
    """Test device_path_from_characteristic_path."""

    _CoInitializeEx(None, COINIT_MULTITHREADED)

    try:
        await assert_mta()
        assert hasattr(allow_sta, "_allowed")
    finally:
        _CoUninitialize()


@pytest.mark.asyncio
async def test_assert_mta_init_sta():
    """Test device_path_from_characteristic_path."""

    _CoInitializeEx(None, COINIT_APARTMENTTHREADED)

    try:
        with pytest.raises(
            BleakError,
            match="Thread is configured for Windows GUI but callbacks are not working.",
        ):
            await assert_mta()
    finally:
        _CoUninitialize()


@pytest.mark.asyncio
async def test_uninitialize_sta():
    """Test device_path_from_characteristic_path."""

    _CoInitializeEx(None, COINIT_APARTMENTTHREADED)
    uninitialize_sta()

    await assert_mta()



import sys
from typing import AsyncGenerator

import pytest
from bumble import data_types
from bumble.core import AdvertisingData, DataType
from bumble.device import Device, DeviceConfiguration
from bumble.gatt import Service
from bumble.hci import Address
from bumble.transport import open_transport
from bumble.transport.common import Transport

from bleak import BleakScanner
from bleak.backends.device import BLEDevice


@pytest.fixture
async def hci_transport(
    request: pytest.FixtureRequest,
) -> AsyncGenerator[Transport, None]:
    """Create a bumble HCI Transport."""
    hci_transport_name: str | None = request.config.getoption("--bleak-hci-transport")
    bluez_vhci_enabled: bool = request.config.getoption("--bleak-bluez-vhci")

    if hci_transport_name is not None and bluez_vhci_enabled:
        raise pytest.UsageError(
            "Cannot use --bleak-hci-transport and --bleak-bluez-vhci together"
        )
    elif bluez_vhci_enabled:
        if sys.platform != "linux":
            pytest.skip("--bleak-bluez-vhci is only supported on Linux")
        from tests.integration.bluez_controller import open_transport_with_bluez_vhci

        async with open_transport_with_bluez_vhci() as hci_transport:
            yield hci_transport
    elif hci_transport_name is not None:
        async with await open_transport(hci_transport_name) as hci_transport:
            yield hci_transport
    else:
        pytest.skip(
            "No HCI transport provided (use --bleak-hci-transport or --bleak-bluez-vhci)"
        )


@pytest.fixture
def bumble_peripheral(hci_transport: Transport) -> Device:
    """
    Create a BLE peripheral device with bumble.
    """
    config = DeviceConfiguration(
        name="Bleak",
        # use random static address to avoid device caching issues, when characteristics change between test runs
        address=Address.generate_static_address(),
        advertising_interval_min=200,
        advertising_interval_max=200,
    )
    return Device.from_config_with_hci(
        config,
        hci_transport.source,
        hci_transport.sink,
    )


def add_default_advertising_data(
    bumble_peripheral: Device,
    additional_adv_data: list[DataType] | None = None,
) -> None:
    """Add default advertising data to bumble peripheral."""
    adv_data: list[DataType] = [
        data_types.Flags(
            AdvertisingData.Flags.LE_GENERAL_DISCOVERABLE_MODE
            | AdvertisingData.Flags.BR_EDR_NOT_SUPPORTED
        ),
        data_types.CompleteLocalName(bumble_peripheral.name),
    ]
    if additional_adv_data:
        adv_data.extend(additional_adv_data)
    bumble_peripheral.advertising_data = bytes(AdvertisingData(adv_data))


async def configure_and_power_on_bumble_peripheral(
    bumble_peripheral: Device,
    additional_adv_data: list[DataType] | None = None,
    services: list[Service] | None = None,
) -> None:
    """Configure and power on the bumble peripheral."""
    add_default_advertising_data(bumble_peripheral, additional_adv_data)
    if services:
        bumble_peripheral.add_services(services)
    await bumble_peripheral.power_on()
    await bumble_peripheral.start_advertising()


async def find_ble_device(bumble_peripheral: Device) -> BLEDevice:
    """Find the BLE device corresponding to the bumble peripheral."""
    device = await BleakScanner.find_device_by_name(bumble_peripheral.name)
    if device is None:
        raise RuntimeError("failed to discover device, is Bumble working?")

    return device



import sys
from collections.abc import Sized
from typing import Protocol

if sys.version_info < (3, 12):
    # mypy and pyright object to Buffer being both ABC and Protocol. Work around
    # by just inheriting from Protocol here. typing_extensions.Buffer is just
    # abc.ABC.
    class BufferProtocol(Protocol):
        def __buffer__(self, flags: int, /) -> memoryview: ...

else:
    from collections.abc import Buffer as BufferProtocol


class SizedBuffer(BufferProtocol, Sized, Protocol):
    """
    Protocol for types that are both Buffer and Sized.

    .. versionadded:: 2.1
    """



import sys

import pytest

# isort: off

if not sys.platform.startswith("linux"):
    pytest.skip("backend only available on Linux", allow_module_level=True)


def test_deprecated_OrPattern_import():
    with pytest.warns(
        DeprecationWarning,
        match="importing OrPattern from bleak.backends.bluezdbus.advertisement_monitor is deprecated",
    ) as recorder:
        from bleak.backends.bluezdbus.advertisement_monitor import (  # noqa: F401
            OrPattern,  # type: ignore[unused-import]
        )

    assert recorder.list[0].filename == __file__


def test_deprecated_OrPatternLike_import():
    with pytest.warns(
        DeprecationWarning,
        match="importing OrPatternLike from bleak.backends.bluezdbus.advertisement_monitor is deprecated",
    ) as recorder:
        from bleak.backends.bluezdbus.advertisement_monitor import (  # noqa: F401
            OrPatternLike,  # type: ignore[unused-import]
        )

    assert recorder.list[0].filename == __file__


def test_deprecated_BlueZDiscoveryFilters_import():
    with pytest.warns(
        DeprecationWarning,
        match="importing BlueZDiscoveryFilters from bleak.backends.bluezdbus.scanner is deprecated",
    ) as recorder:
        from bleak.backends.bluezdbus.scanner import (  # noqa: F401
            BlueZDiscoveryFilters,  # type: ignore[unused-import]
        )
    assert recorder.list[0].filename == __file__


def test_deprecated_BlueZScannerArgs_import():
    with pytest.warns(
        DeprecationWarning,
        match="importing BlueZScannerArgs from bleak.backends.bluezdbus.scanner is deprecated",
    ) as recorder:
        from bleak.backends.bluezdbus.scanner import (  # noqa: F401
            BlueZScannerArgs,  # type: ignore[unused-import]
        )

    assert recorder.list[0].filename == __file__



import sys

import pytest

# isort: off

if not sys.platform.startswith("win"):
    pytest.skip("backend only available on windows", allow_module_level=True)


def test_deprecated_WinRTClientArgs_import():
    with pytest.warns(
        DeprecationWarning,
        match="importing WinRTClientArgs from bleak.backends.winrt.client is deprecated, use bleak.args.winrt instead",
    ) as recorder:
        from bleak.backends.winrt.client import (  # noqa: F401
            WinRTClientArgs,  # type: ignore[unused-import]
        )

    assert recorder.list[0].filename == __file__



"""
-----------------------
BlueZ backend arguments
-----------------------
"""

from typing import NamedTuple, TypedDict, Union

from bleak.assigned_numbers import AdvertisementDataType


class BlueZDiscoveryFilters(TypedDict, total=False):
    """
    Dictionary of arguments for the ``org.bluez.Adapter1.SetDiscoveryFilter``
    D-Bus method.

    https://github.com/bluez/bluez/blob/master/doc/org.bluez.Adapter.rst#void-setdiscoveryfilterdict-filter
    """

    UUIDs: list[str]
    """
    Filter by service UUIDs, empty means match _any_ UUID.

    Normally, the ``service_uuids`` argument of :class:`bleak.BleakScanner`
    is used instead.
    """
    RSSI: int
    """
    RSSI threshold value.
    """
    Pathloss: int
    """
    Pathloss threshold value.
    """
    Transport: str
    """
    Transport parameter determines the type of scan.

    This should not be used since it is required to be set to ``"le"``.
    """
    DuplicateData: bool
    """
    Disables duplicate detection of advertisement data.

    This does not affect the ``Filter Duplicates`` parameter of the ``LE Set Scan Enable``
    HCI command to the Bluetooth adapter!

    Although the default value for BlueZ is ``True``, Bleak sets this to ``False`` by default.
    """
    Discoverable: bool
    """
    Make adapter discoverable while discovering,
    if the adapter is already discoverable setting
    this filter won't do anything.
    """
    Pattern: str
    """
    Discover devices where the pattern matches
    either the prefix of the address or
    device name which is convenient way to limited
    the number of device objects created during a
    discovery.
    """


class OrPattern(NamedTuple):
    """
    BlueZ advertisement monitor or-pattern.

    https://github.com/bluez/bluez/blob/master/doc/org.bluez.AdvertisementMonitor.rst#arrayuint8-uint8-arraybyte-patterns-read-only-optional
    """

    start_position: int
    ad_data_type: AdvertisementDataType
    content_of_pattern: bytes


# Windows has a similar structure, so we allow generic tuple for cross-platform compatibility
OrPatternLike = Union[OrPattern, tuple[int, AdvertisementDataType, bytes]]


class BlueZScannerArgs(TypedDict, total=False):
    """
    :class:`BleakScanner` args that are specific to the BlueZ backend.
    """

    filters: BlueZDiscoveryFilters
    """
    Filters to pass to the adapter SetDiscoveryFilter D-Bus method.

    Only used for active scanning.
    """

    or_patterns: list[OrPatternLike]
    """
    Or patterns to pass to the AdvertisementMonitor1 D-Bus interface.

    Only used for passive scanning.
    """


class BlueZNotifyArgs(TypedDict, total=False):
    """
    :meth:`bleak.BleakClient.start_notify` method args that are specific to the
    BlueZ backend.

    .. versionadded:: 2.1
    """

    use_start_notify: bool
    """
    If true, use the "StartNotify" D-Bus method instead of "AcquireNotify" to
    subscribe to notifications.

    This is needed in rare cases to work around BlueZ quirks. For example, some
    peripherals may send notifications immediately after writing to the CCCD
    descriptor, before the write response is sent. In this case, "AcquireNotify"
    will miss the notification, whereas "StartNotify" will work correctly.
    """
